import threading
import io
import pathlib
import tarfile
import typing
from collections import defaultdict
from collections.abc import Iterable, Iterator
from typing import Union

import xxhash


def _defaultdict_recursive():
    return defaultdict(_defaultdict_recursive)


class ProjFsNode:
    def __init__(self, dirty: bool):
        self._dirty = dirty

    def __getitem__(self, item):
        raise NotImplementedError

    @property
    def dirty(self):
        return self._dirty

    def set_node(
        self,
        fn: str,
        value: "ProjFsNode",
    ) -> "ProjFsNode":
        raise NotImplementedError

    def get_hash(self):
        if not hasattr(self, "_hash"):
            hasher = xxhash.xxh64()
            hasher.update(self._name)
            hasher.update(self._content)
            self._hash = hasher.hexdigest()
        return self._hash

    def get_size(self, recursive: bool = False) -> int:
        """
        Get the size of this node in bytes.
        
        Args:
            recursive: If True and this is a directory, include all children recursively.
                      If False, only count this node's direct content.
        
        Returns:
            Size in bytes
        """
        raise NotImplementedError

    def __hash__(self):
        return hash(self.get_hash())


class ProjectDir(ProjFsNode):
    """
    Represents a directory. It is a persistent data structure (mutations to
    it return a new instance).
    """

    _contents: dict[str, ProjFsNode]

    def __init__(
        self,
        contents: dict[str, ProjFsNode],
        dirty: bool = False,
    ) -> None:
        super().__init__(dirty=dirty)
        self._contents = contents

    @classmethod
    def construct_with_files(
        cls,
        files: Iterable[tuple[str, bytes | str]],
    ) -> "ProjectDir":
        """Creates a new ProjectDir with the given files added. Works with nested paths."""
        # data structure to store the file hierarchy
        root = _defaultdict_recursive()

        for path, content in files:
            parts = path.split("/")

            # navigate through the file hierarchy and set the content
            d = root
            for part in parts[:-1]:
                d = d[part]
            d[parts[-1]] = content

        # convert data structure to ProjectDir
        def create_dir(d):
            contents = {}
            for name, val in d.items():
                if isinstance(val, dict):
                    contents[name] = create_dir(
                        val,
                    )  # recursive call for nested dirs
                else:
                    # base case: create a ProjectFile for each file content
                    contents[name] = ProjectFile(name, val)
            return ProjectDir(contents)

        return create_dir(root)

    @classmethod
    def construct_with_one_file(
        cls,
        path: str,
        content: bytes | str,
    ) -> "ProjectDir":
        """Constructs a ProjectDir with a single file."""
        return cls.construct_with_files([(path, content)])

    def __getitem__(self, fn: str) -> Union["ProjectDir", "ProjectFile"]:
        if "/" in fn:
            d, rest = split_fn(fn)
            return self._contents[d][rest]
        return self._contents[fn]

    def __contains__(self, fn: str) -> bool:
        if "/" in fn:
            d, rest = split_fn(fn)
            return d in self._contents and rest in self._contents[d]
        return fn in self._contents

    def __iter__(self):
        yield from self._contents.values()

    def set_node(
        self,
        fn: str,
        value: ProjFsNode,
        clean_value_ok: bool = False,
    ) -> "ProjectDir":
        """
        Returns a new ProjectDir with the given file set to the given value.

        :param clean_value_ok: By default attempting to set a node with a value
            marked as clean will switch it to a dirty. If it is a directory
            it will raise an error. Set false if setting a clean value is intentional.
        """
        if not value.dirty and not clean_value_ok:
            if isinstance(value, ProjectFile):
                value = ProjectFile(value.name, value.content, dirty=True)
            else:
                msg = (
                    "Attempting to set a node with a clean value. Expected it "
                    "to now be dirty unless clean_value_ok=True"
                )
                raise ValueError(msg)
        new_contents = self._contents.copy()
        first, rest = split_fn(fn)
        if rest:
            new_contents[first] = self._contents[first].set_node(rest, value)
        else:
            if isinstance(value, ProjectFile) and value.name != first:
                msg = "Node value name does not match first"
                raise ValueError(msg)
            new_contents[first] = value
        return ProjectDir(new_contents, dirty=value.dirty)

    def set_all(
        self,
        new_files: "ProjectDir",
        dirty: bool = True,
    ) -> "ProjectDir":
        """
        Sets all files and directories from another ProjectDir into this one
        recursively, overwriting existing content.
        """
        new_contents = self._contents.copy()

        for name, node in new_files._contents.items():
            if (
                name in new_contents
                and isinstance(node, ProjectDir)
                and isinstance(new_contents[name], ProjectDir)
            ):
                new_contents[name] = new_contents[name].set_all(node)
            else:
                if not dirty and node.dirty:
                    msg = (
                        "Attempting to set a node with a dirty value. "
                        "Expected it to now be clean unless dirty=True"
                    )
                    raise ValueError(msg)
                new_contents[name] = node

        return ProjectDir(new_contents, dirty=dirty)

    def set_file_contents(
        self,
        path: str,
        content: bytes | str,
        dirty: bool = True,
    ) -> "ProjectDir":
        """Returns a new ProjectDir with the given file set to the given value."""
        fn = path.split("/")[-1]
        new_file = ProjectFile(fn, content, dirty=dirty)
        return self.set_node(path, new_file)

    def get_only_file(self, only_consider_dirty: bool = False) -> "ProjectFile":
        """
        Gets the single file in this directory (or descendent files). If
        there is not exactly one file, raises an AssertionError.
        """
        files = list(
            self.walk(include_dirs=False, only_consider_dirty=only_consider_dirty),
        )
        if len(files) != 1:
            msg = f"Expected exactly one file, got {len(files)}"
            raise AssertionError(msg)
        return files[0][1]

    def walk(
        self,
        include_dirs: bool = False,
        only_consider_dirty: bool = False,
    ) -> Iterator[tuple[str, ProjFsNode]]:
        """Walks the directory tree, yielding (path, node) pairs."""
        for name, node in self._contents.items():
            if only_consider_dirty and not node.dirty:
                continue
            if isinstance(node, ProjectDir):
                if include_dirs:
                    yield name, node
                yield from (
                    (f"{name}/{path}", node)
                    for path, node in node.walk(
                        include_dirs=include_dirs,
                        only_consider_dirty=only_consider_dirty,
                    )
                )
            else:
                yield name, node

    def pretty_print(self, indent=0, recursive=True, file: typing.IO | None = None):
        """Pretty print the directory contents."""
        for name, node in self._contents.items():
            print(f"{' ' * indent}{name}{' (Dirty)' if node.dirty else ''}", file=file)
            if recursive and isinstance(node, ProjectDir):
                node.pretty_print(indent=indent + 2, recursive=recursive, file=file)

    def pretty_str(self):
        file_str = io.StringIO()
        self.pretty_print(file=file_str)
        return file_str.getvalue()

    def contains_only_leafs(self) -> bool:
        return all(isinstance(v, ProjectFile) for v in self._contents.values())

    def _add_to_tar(self, tar, path=""):
        """Recursive function to add ProjectFile or ProjectDir to tar."""
        for name, node in self._contents.items():
            if isinstance(node, ProjectFile):
                tarinfo = tarfile.TarInfo(name=path + node.name)
                tarinfo.size = len(node.content)
                tar.addfile(tarinfo, io.BytesIO(node.content))
            elif isinstance(node, ProjectDir):
                node._add_to_tar(tar, path + name + "/")
            else:
                msg = f"Unexpected type {type(node)}"
                raise TypeError(msg)

    def convert_to_tar(self, gzip: bool = False) -> bytes:
        """Converts the directory to a tarball."""
        tar_data = io.BytesIO()
        with tarfile.open(
            fileobj=tar_data,
            mode="w:gz" if gzip else "w",
        ) as tar:
            self._add_to_tar(tar)
        return tar_data.getvalue()

    @classmethod
    def from_path(cls, path: str | pathlib.Path) -> "ProjectDir":
        if isinstance(path, str):
            path = pathlib.Path(path)

        dir_content = {}
        for item in path.iterdir():
            if item.is_file():
                with item.open("rb") as f:
                    content = f.read()
                    dir_content[item.name] = ProjectFile(
                        name=item.name,
                        content=content,
                    )
            elif item.is_dir():
                dir_content[item.name] = cls.from_path(item)
        return cls(contents=dir_content)

    def expand_glob(self, glob: str) -> Iterator[str]:
        """Expands a glob pattern to a list of files."""
        if "*" not in glob and glob in self:
            yield glob
            return
        if "**" in glob:
            msg = "Recursive glob not implemented"
            raise NotImplementedError(msg)
        if "*" in glob:
            msg = "Non-recursive glob not implemented"
            raise NotImplementedError(msg)

    def __eq__(self, other):
        if not isinstance(other, ProjectDir):
            return False
        return self._contents == other._contents

    def get_hash(self):
        if (not hasattr(self, "_hash")) or self._hash is None:
            hasher = xxhash.xxh64()
            for name, node in sorted(self._contents.items()):
                hasher.update(name)
                hasher.update(node.get_hash())
            self._hash = hasher.hexdigest()
        return self._hash

    def get_size(self, recursive: bool = False) -> int:
        """
        Get the size of this directory in bytes.
        
        Args:
            recursive: If True, include all children recursively.
                      If False, return 0 (directories have no direct content).
        
        Returns:
            Size in bytes
        """
        if not recursive:
            return 0  # Directories themselves have no size
        
        total_size = 0
        for node in self._contents.values():
            total_size += node.get_size(recursive=True)
        return total_size

    def __hash__(self):
        return hash(self.get_hash())

    def __repr__(self):
        return f"ProjectDir({len(self._contents)} children; {self.get_hash()}{', dirty' if self.dirty else ''})"


class ProjectFile(ProjFsNode):
    """Immutable file."""

    def __init__(self, name: str, content: bytes | str, dirty: bool = False) -> None:
        super().__init__(dirty=dirty)
        self._name = name
        if isinstance(content, str):
            self._content = content.encode("utf-8")
        elif isinstance(content, bytes):
            self._content = content
        else:
            msg = f"Unexpected type {type(content)}"
            raise TypeError(msg)

    @property
    def name(self):
        return self._name

    @property
    def content(self) -> bytes:
        return self._content

    @property
    def content_str(self) -> str:
        return self._content.decode("utf-8")

    def __getitem__(self, item):
        msg = "Cannot index into a file"
        raise RuntimeError(msg)

    def __eq__(self, other):
        if not isinstance(other, ProjectFile):
            return False
        return self._name == other._name and self._content == other._content

    def get_hash(self):
        if (not hasattr(self, "_hash")) or self._hash is None:
            hasher = xxhash.xxh64()
            hasher.update(self._name)
            hasher.update(self._content)
            self._hash = hasher.hexdigest()
        return self._hash

    def get_size(self, recursive: bool = False) -> int:
        """
        Get the size of this file in bytes.
        
        Args:
            recursive: Ignored for files (files are always leaf nodes).
        
        Returns:
            Size of file content in bytes
        """
        return len(self._content)


def split_fn(fn) -> tuple[str, str]:
    if "/" in fn:
        d, rest = fn.split("/", 1)
        return d, rest
    return fn, ""


def create_project_dir_from_docker_img(
    image_name: str,
    path: str = "/",
    tag: str = "latest",
) -> ProjectDir:
    """
    Extract a specific path from a Docker image and return as ProjectDir.
    
    Args:
        image_name: Docker image name (e.g., "nginx", "python", "nginx:latest")
        path: Path within the image to extract (default: root "/")
        tag: Image tag (default: "latest") - ignored if image_name already contains a tag
    
    Returns:
        ProjectDir containing the extracted filesystem
        
    Raises:
        subprocess.CalledProcessError: If docker commands fail
        FileNotFoundError: If the specified path doesn't exist in the image
        
    Example:
        # Extract entire filesystem
        fs = create_project_dir_from_docker_img("nginx:latest")
        
        # Extract specific directory
        app = create_project_dir_from_docker_img("my-app", "/app/src", "v1.0")
    """
    import subprocess
    import tempfile
    import shutil
    
    # Handle image name with or without tag
    if ":" in image_name:
        # Image name already includes tag, use as-is
        full_image_name = image_name
    else:
        # Add tag to image name
        full_image_name = f"{image_name}:{tag}"
    
    # Create temporary container (doesn't start it)
    # Use a dummy command in case the image doesn't have a default CMD
    try:
        result = subprocess.run([
            "docker", "create", full_image_name, "sleep", "0.01"
        ], capture_output=True, text=True, check=True)
    except subprocess.CalledProcessError as e:
        raise subprocess.CalledProcessError(
            e.returncode, 
            e.cmd, 
            f"Failed to create container from image {full_image_name}. "
            f"Make sure the image exists and Docker is running.\n{e.stderr}"
        ) from e
    
    container_id = result.stdout.strip()
    
    try:
        # Create temporary directory for extraction
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = pathlib.Path(temp_dir)
            
            # Copy specific path from container
            try:
                subprocess.run([
                    "docker", "cp", 
                    f"{container_id}:{path}", 
                    str(temp_path)
                ], check=True, capture_output=True)
            except subprocess.CalledProcessError as e:
                raise FileNotFoundError(
                    f"Path '{path}' not found in image {full_image_name}"
                ) from e
            
            # The copied content will be at temp_path/basename(path)
            # Handle special case where path is root "/"
            if path == "/":
                # docker cp container:/ copies the contents, not the root dir itself
                return ProjectDir.from_path(temp_path)
            else:
                # docker cp container:/some/path creates temp_dir/path_basename
                copied_name = pathlib.Path(path).name
                extracted_path = temp_path / copied_name
                
                if not extracted_path.exists():
                    raise FileNotFoundError(
                        f"Expected extracted path {extracted_path} not found"
                    )
                
                return ProjectDir.from_path(extracted_path)
    
    finally:
        # Clean up container
        try:
            subprocess.run([
                "docker", "rm", container_id
            ], check=True, capture_output=True)
        except subprocess.CalledProcessError:
            # Log warning but don't fail - cleanup is best effort
            pass



import multiprocessing
import random
from collections.abc import Callable
from pathlib import Path

import lmwrapper.caching
import pandas as pd
from lmwrapper.huggingface_wrapper import get_huggingface_lm
from lmwrapper.openai_wrapper import OpenAiModelNames, get_open_ai_lm

from synthegrator.code_problems import CodeProblem
from synthegrator.code_solver import (
    LmBasedSolver,
    LmCodeSolverAutoRegressive,
    ResponseParserSingleEdit,
)
from synthegrator.df_converters import solution_evals_to_df
from synthegrator.execution_threading import solve_and_evaluate_problems
from synthegrator.problem_rendering import (
    PromptRendererSingleEditGeneric,
)
from synthegrator.problem_rendering_insertion_tags import (
    TaggedEditRenderer,
    TaggedEditResponseParser,
)
from synthegrator.synthdatasets.human_eval import yield_human_eval

random.seed(0)

try:
    import torch
except ImportError:
    torch = None

cur_file = Path(__file__).parent.absolute()


def get_experiment_solver_constructors(
    open_ai_models: bool = True,
    hf_models: bool = torch is not None and torch.cuda.is_available(),
    use_edit_renders: bool = False,
) -> dict[str, Callable[[], LmBasedSolver]]:
    """
    Returns a dictionary of solver constructors for all the models
    we want to run in this experiment. The dictionary keys are
    model names. The values are lambda functions that construct
    the solvers. We don't set the values as the actual solvers
    as they might load large models that we don't/can't keep
    all in memory together. So instead by using the lambda
    we can lazily construct the solvers only when they
    are needed.
    """
    solvers = {}
    if use_edit_renders:
        renderer = TaggedEditRenderer()
        response_parser = TaggedEditResponseParser()
    else:
        renderer = PromptRendererSingleEditGeneric()
        response_parser = ResponseParserSingleEdit()

    def construct(lm):
        return LmCodeSolverAutoRegressive(
            lm,
            prompt_renderer=renderer,
            response_parser=response_parser,
            # snap_indent=4,
            include_lm_response=True,
            cache_lm=True,
        )

    if open_ai_models:
        model_names = [
            OpenAiModelNames.gpt_3_5_turbo_instruct,
            OpenAiModelNames.gpt_3_5_turbo,
            #OpenAiModelNames.gpt_4_turbo,
            OpenAiModelNames.gpt_4_1_nano,
            OpenAiModelNames.gpt_4_1,
        ]
        solvers |= {
            # Note: Closure needs an arg so it gets captured
            #   without changing later in the loop.
            str(model_name): (lambda model_name=model_name: construct(
                get_open_ai_lm(
                    model_name,
                    retry_on_rate_limit=True,
                )
            ))
            for model_name in model_names
        }
    if hf_models:
        model_names = [
            "Salesforce/codegen-350M-multi",
            "mistralai/Mistral-7B-v0.1",
            "Qwen/Qwen2.5-Coder-0.5B",
        ]
        solvers |= {
            model_name: (lambda model_name=model_name: construct(
                get_huggingface_lm(
                    model_name,
                    trust_remote_code=True,
                    precision=torch.float16,
                )
            ))
            for model_name in model_names
        }
    return solvers


def run_for_problems_and_serialize(
    problems: list[CodeProblem],
    task_name: str,
    output_directory: Path = cur_file / "results",
    max_limit: int = 500,
    pickle_gzip_whole_solution_eval: bool = False,
    read_file_if_exists: bool = False,
    solver_constructors: dict[
        str,
        Callable[[], LmBasedSolver],
    ] = get_experiment_solver_constructors(),
) -> pd.DataFrame:
    if problems is None:
        raise ValueError("problems parameter cannot be None")

    problems_sample = None

    def get_problems_sample():
        nonlocal problems_sample
        if problems_sample is not None:
            return problems_sample
        if len(problems) > max_limit:
            problems_sample = random.sample(problems, max_limit)
        else:
            problems_sample = problems
        return problems_sample

    thread_goal = max(multiprocessing.cpu_count() // 3, 1)
    print("thread_goal: ", thread_goal)
    lmwrapper.caching.set_cache_dir(
        Path(__file__).parent.absolute() / "lm_wrapper_cache",
    )
    dfs = []
    for solver_name, solver_constructor in solver_constructors.items():
        dest_file = output_directory / f"results_{task_name}_{solver_name.replace('/', '_')}.parquet"
        if read_file_if_exists and dest_file.exists():
            dfs.append(pd.read_parquet(dest_file))
            continue
        print(f"Constructing new solver {solver_name}...")
        solver = solver_constructor()
        print(f"Running {solver_name}...")
        evals = solve_and_evaluate_problems(
            solver=solver,
            problems=get_problems_sample(),
            max_threads_solve=max(min(thread_goal, 3), 1),
            max_threads_eval=thread_goal,
        )
        print(f"Converted {solver_name} to df...")
        new_df = solution_evals_to_df(
            evals,
            solver_key=solver_name,
            pickle_gzip_whole_solution_eval=pickle_gzip_whole_solution_eval,
        )
        print("Converted to df")
        new_df["task_name"] = task_name
        new_df["task_problem_count"] = len(problems)
        for eval in evals:
            assert eval.solution.problem is not None
        new_df.to_parquet(dest_file)
        print("Saved to parquet.", dest_file)
        dfs.append(new_df)
        del solver
    df = pd.concat(dfs)
    scores = df.groupby("solver_key").main_metric__is_success.mean()
    print(scores)
    return df


def get_human_eval_results(
    output_directory: Path = cur_file / "results",
    read_file_if_exists: bool = False,
    max_limit: int = 500,
) -> pd.DataFrame:
    print("Gathering all problems...")
    problems = list(yield_human_eval())
    return run_for_problems_and_serialize(
        problems,
        "human_eval",
        output_directory=output_directory,
        read_file_if_exists=read_file_if_exists,
        max_limit=max_limit,
    )


def main():
    get_human_eval_results()


if __name__ == "__main__":
    main()



import re
import textwrap
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path

import libcst as cst
from tree_sitter import Language, Node, Tree
#from tree_sitter_languages import get_language, get_parser
from tree_sitter_language_pack import get_language, get_parser

from synthegrator.lang_specs.lang_spec import (
    LangSpec,
    LsFunctionDef,
    PrefixComment,
    PythonFunctionParser,
)
from synthegrator.lang_specs.lang_util import truncate_when_decreases_indent
from synthegrator.sandboxing import PY_DEFAULT_DOCKER_ENV, DockerExecutionContext
from synthegrator.util import lazy_prop


def assert_spec_is_python(spec: LangSpec):
    if not isinstance(spec, PythonLangSpec):
        msg = "Only python language supported"
        raise TypeError(msg)


class PythonLangSpec(LangSpec):
    @staticmethod
    def get_tree_sitter_lang_name() -> str:
        return "python"

    def truncate_function_completion(self, completion: str) -> str:
        return truncate_when_decreases_indent(completion)

    def get_method_name(self, method_src) -> str:
        import ast

        tree = ast.parse(method_src)
        if not isinstance(tree, ast.Module):
            msg = "Tree must be of ast.Module type"
            raise TypeError(msg)
        if len(tree.body) != 1:
            msg = "More than one tree body"
            raise ValueError(msg)
        func_def = tree.body[0]
        if not isinstance(func_def, ast.FunctionDef):
            msg = "Function definition must be of ast.FunctionDef type"
            raise TypeError(msg)
        return func_def.name

    def get_lang_md_name(self) -> str:
        return "python"

    @staticmethod
    def split_before_last_method_name(code: str) -> tuple[str, str]:
        # method dec when line starts with def and ends with and ends with :
        lines = code.split("\n")
        for i, line in enumerate(reversed(lines)):
            if line.startswith("def") and line.endswith(":"):
                before = "\n".join(lines[: -i - 1])
                if len(lines[: -i - 1]) > 0:
                    before += "\n"
                method = "\n".join(lines[-i - 1 :])
                return before, method
        msg = "No method declaration found"
        raise RuntimeError(msg)

    @staticmethod
    def get_signature_text(method: str) -> str:
        module = cst.parse_module(method)
        func_def = module.body[0]
        if not isinstance(func_def, cst.FunctionDef):
            msg = "Function definition not of cst.FunctionDef type"
            raise TypeError(msg)

        body_code = module.code_for_node(func_def.body)
        # return everything before the body
        return method[: method.index(body_code)].strip()

    @staticmethod
    def get_default_docker_env() -> DockerExecutionContext:
        return PY_DEFAULT_DOCKER_ENV

    @classmethod
    def find_functions(
        cls,
        src: str | Tree,
        include_class_functions: bool = True,
        include_nested_functions: bool = True,
    ) -> Iterable["LsFunctionDefPython"]:
        """
        Finds all functions in the source code.
        :param include_class_functions: If true, will include functions defined inside classes.
        :param include_nested_functions: If true, will include functions defined inside other functions.
        """
        if not include_nested_functions:
            raise NotImplementedError
        if not include_class_functions:
            raise NotImplementedError
        tree = cls.get_tree_sitter_tree(src) if isinstance(src, str) else src
        captures = _py_function_def_query().captures(tree.root_node)
        for node in captures.get('function', []):
            yield LsFunctionDefPython(node=node, tree=tree)

    @classmethod
    def get_prefix_comments(cls, src: str | Tree) -> Iterable[PrefixComment]:
        tree = cls.get_tree_sitter_tree(src) if isinstance(src, str) else src
        captures = _py_prefix_comments_query().captures(tree.root_node)
        if len(captures) == 0:
            return
        for captured_node in captures['docstring']:
            yield PrefixComment(
                doc_string_node=captured_node,
                function_node=captured_node.parent.parent.parent,
            )

    @staticmethod
    def get_tree_sitter_lang() -> Language:
        return PY_LANG_TS


@lazy_prop
def _py_prefix_comments_query():
    return PY_LANG_TS.query(
        """
        (function_definition
          (block .
            (expression_statement
              (string) @docstring)))
        """,
    )


@lazy_prop
def _py_function_def_query():
    return PY_LANG_TS.query(
        """
        (function_definition) @function
        """
    )


@lazy_prop
def _py_docstring_query_in_func():
    return PY_LANG_TS.query(
        """
        (expression_statement
          (string) @docstring
          (#not-has-children? @docstring)
        )
        """
    )


@lazy_prop
def _docstring_pattern():
    return re.compile(r'^\s*("""[\s\S]*?"""|\'\'\'[\s\S]*?\'\'\')')



class PythonModuleParser:
    def __init__(self, module_source: str) -> None:
        self.language = get_language("python")
        self.parser = get_parser("python")
        self.module_source = module_source
        self.function_without_doc = None
        self.function_definition_with_doc = None
        self.function_definition_without_doc = None
        self.body_span_begin = -1
        self.body_span_end = -1
        self.has_docstring = False
        self.parse_source()


PY_LANG_TS = get_language("python")


def _main():
    """Test"""
    langspec_py = PythonLangSpec()
    source_code = Path(__file__).read_text()
    tree = langspec_py.get_tree_sitter_tree(source_code)
    root_node = tree.root_node

    if root_node.type != "module":
        msg = "Module not provided."
        raise Exception(msg)

    for node in root_node.children:
        if node.type in {"decorated_definition", "function_definition"}:
            func_parse = PythonFunctionParser(node.text.decode())
            print((func_parse.function_definition_with_doc or "") + "\n    ...")
        if node.type == "class_definition":
            print(
                source_code[node.children[0].start_byte : node.children[-2].end_byte],
            )
            for cnode in node.children[-1].children:
                if cnode.type in {
                    "decorated_definition",
                    "function_definition",
                }:
                    func_parse = PythonFunctionParser(cnode.text.decode())
                    print(func_parse.function_definition_with_doc + "\n    ...")
                if cnode.type == "comment":
                    print(cnode.text.decode())


@dataclass(frozen=True)
class LsFunctionDefPython(LsFunctionDef):
    """
    A wrapper around a function definition with useful functions common
    between languages.

    Note: this python specific. When expand language specs will need to
    make subclasses following this API.

    For reference, here is the treesitter grammar for function definitions:
    https://github.com/tree-sitter/tree-sitter-python/blob/c01fb4e/src/grammar.json#L1582-L1671
    """

    def get_body_node(self) -> Node:
        """Returns the node that corresponds to the body of the function"""
        return self.node.children[-1]

    # @frozen_method_cache
    def get_body_src(
        self,
        include_body_indent: bool = True,
        include_trailing_new_line: bool = True,
        include_prefix_comment: bool = True,
    ) -> str:
        """
        Source of the method (potentially including docstring). The body node
        won't include the body indent (the leading whitespace). Instead that is part
        of the function declaration. The `include_body_indent` can be used to grab it
        
        :param include_prefix_comment: If False, excludes the prefix comment (docstring in Python) from the body
        """
        declaration_src = self.get_declaration_str(
            include_body_indent=not include_body_indent,
        )
        full_src = self.get_full_function_src()
        src = full_src[len(declaration_src) :]
        
        # If we need to exclude the prefix comment (docstring), remove it
        if not include_prefix_comment:
            prefix_comment = self.get_prefix_comment_str()
            if prefix_comment is not None:
                # Use the shared regex pattern to find and remove the docstring
                match = _docstring_pattern().match(src)
                if match:
                    docstring_with_indent = match.group(0)
                    # Remove the docstring and any trailing newlines
                    src = src[len(docstring_with_indent):].lstrip('\n')
                    # Preserve the original indentation for the remaining code
                    if src and not include_body_indent:
                        # If we don't want body indent, keep as is
                        pass
                    # If there's remaining content, we're good
                    # If no remaining content, src will be empty which is correct
        
        if include_trailing_new_line:
            src += "\n"
        return src

    def get_preceding_src_code(self):
        return self.tree.root_node.text[: self.node.start_byte].decode()

    def get_src_code_after(self):
        """Returns the source code after the function"""
        return self.tree.root_node.text[self.node.end_byte :].decode()

    # @frozen_method_cache
    def get_function_name(self) -> str:
        """Returns the name identifier of the function"""
        has_async = self.node.children[0].type == "async"
        if has_async:
            # async, def, name
            return self.node.children[2].text.decode()
        # def, name
        return self.node.children[1].text.decode()

    # @frozen_method_cache
    def get_declaration_str(self, include_body_indent: bool = False) -> str:
        r"""
        Returns the full declaration of the function (so the `def`, signature, args, and `:`

        :param include_body_indent: If false the return might have leading ident.
            For example "def foo():\n    pass" would return "def foo():\n    "
            with the indent. If false would return "def foo():\n"
        """
        body_node = self.get_body_node()
        start_byte = self.node.start_byte
        end_byte = body_node.start_byte
        text = self.tree.root_node.text[start_byte:end_byte].decode()
        if include_body_indent:
            return text
        last_colon_idx = text.rfind(":")
        has_newline = "\n" in text[last_colon_idx:]
        if has_newline:
            return text[: last_colon_idx + 2]
        return text[: last_colon_idx + 1]

    # @frozen_method_cache
    def get_full_function_src(self, zero_base_indent: bool = False) -> str:
        """Returns the entire source code of the function, optionally without leading indentation"""
        src = self.node.text.decode()
        if not zero_base_indent:
            return src
        return textwrap.dedent(src)

    # @frozen_method_cache
    def get_prefix_comment_str(self, as_plain_text: bool = False) -> str | None:
        """
        Gets the prefix comment (a docstring in Python)

        :param as_plain_text: Normally will include the language specific annotations
            (such as the triple quotes in Python). If true, will return the dedented plain text.

        Body spec (a bit tricky, so we will just do it with string regex):
        https://github.com/tree-sitter/tree-sitter-python/blob/c01fb4e3/src/grammar.json#L2178-L2213
        """
        body_src = self.get_body_src()
        if len(body_src) == 0:
            return None

        # Use the shared regex pattern for docstring detection
        match = _docstring_pattern().match(body_src)
        if match:
            docstring = match.group(1)
            if as_plain_text:
                # remove quotes and dedent
                unquoted = docstring[3:-3]
                return textwrap.dedent(unquoted)
            return docstring

        # Return None if no docstring was found
        return None


if __name__ == "__main__":
    _main()



"""
Tests for dypybench problems. The 'x' in the module name is to make it run
late after other tests (since it is slow)
"""

import dataclasses
import os
from pathlib import Path

import pytest
from lmwrapper.openai_wrapper import get_open_ai_lm, OpenAiModelNames

from synthegrator.code_problems import DiscoveredTestsuite
from synthegrator.code_solver import DummyCodeSolverAutoRegressive, LmCodeSolverAutoRegressive
from synthegrator.problem_rendering import render_single_edit_codeproblem_autoregressive
from synthegrator.sandboxing import run_on_docker
from synthegrator.solution_eval import (
    apply_solution,
    evaluate_code_problem_execution,
    exact_match_metric,
    is_solution_syntax_likely_valid,
    parse_junit_test_cases,
)
from synthegrator.synthdatasets.dypybench import (
    _default_projects_root,
    _project_docker_envs,
    produce_dy_py_bench_problem,
    yield_dypybench,
)
from synthegrator.transformation_spec import StsEditable, StsSelectorLineSpan
from synthegrator.util import MultilineString

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


@pytest.fixture
def example_dypy_prob():
    fixture = Path(__file__).parent / "fixtures/document.py.fixture"

    function_w_doc = """def process_encoding(self, encoding: None | str = None) -> str:
    \"\"\"Process explicitly defined encoding or auto-detect it.

    If encoding is explicitly defined, ensure it is a valid encoding the python
    can deal with. If encoding is not specified, auto-detect it.

    Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
    \"\"\"
    if encoding:
        return unicodec.normalize_encoding_name(encoding)
    return unicodec.detect_content_encoding(
        self.get_body_chunk() or b"",
        content_type_header=(
            self.headers.get("Content-Type", None) if self.headers else None
        ),
        markup="xml" if self.document_type == "xml" else "html",
    )"""

    import tempfile

    with tempfile.TemporaryDirectory() as tmpdir:
        # Make a project1/grab/document.py file in the tmpdir with contents of fixture
        Path(tmpdir).joinpath("project1/grab").mkdir(parents=True, exist_ok=True)
        Path(tmpdir).joinpath("project1/grab/document.py").write_text(
            fixture.read_text(),
        )

        problem_w_doc = produce_dy_py_bench_problem(
            task_id=str(1),
            path="grab/document.py",
            identifier="Document.process_encoding",
            func_begin=134,
            func_end=150,
            function=function_w_doc,
            project_id=str(1),
            projects_root=tmpdir,
            random_line_seed=2,
        )

        yield problem_w_doc, function_w_doc


@pytest.fixture
def example_dypy_prob_stripdoc():
    fixture = Path(__file__).parent / "fixtures/document.py.fixture"

    function_w_doc = """def process_encoding(self, encoding: None | str = None) -> str:
    \"\"\"Process explicitly defined encoding or auto-detect it.

    If encoding is explicitly defined, ensure it is a valid encoding the python
    can deal with. If encoding is not specified, auto-detect it.

    Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
    \"\"\"
    if encoding:
        return unicodec.normalize_encoding_name(encoding)
    return unicodec.detect_content_encoding(
        self.get_body_chunk() or b"",
        content_type_header=(
            self.headers.get("Content-Type", None) if self.headers else None
        ),
        markup="xml" if self.document_type == "xml" else "html",
    )"""

    import tempfile

    with tempfile.TemporaryDirectory() as tmpdir:
        # Make a project1/grab/document.py file in the tmpdir with contents of fixture
        Path(tmpdir).joinpath("project1/grab").mkdir(parents=True, exist_ok=True)
        Path(tmpdir).joinpath("project1/grab/document.py").write_text(
            fixture.read_text(),
        )

        problem_w_doc = produce_dy_py_bench_problem(
            task_id=str(1),
            path="grab/document.py",
            identifier="Document.process_encoding",
            func_begin=134,
            func_end=150,
            function=function_w_doc,
            project_id=str(1),
            projects_root=tmpdir,
            strip_docstring=True,
            random_line_seed=2,
        )

        yield problem_w_doc, function_w_doc


@pytest.fixture
def example_dypy_prob_stripdoc_fixedline():
    fixture = Path(__file__).parent / "fixtures/document.py.fixture"

    function_w_doc = """def process_encoding(self, encoding: None | str = None) -> str:
    \"\"\"Process explicitly defined encoding or auto-detect it.

    If encoding is explicitly defined, ensure it is a valid encoding the python
    can deal with. If encoding is not specified, auto-detect it.

    Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
    \"\"\"
    if encoding:
        return unicodec.normalize_encoding_name(encoding)
    return unicodec.detect_content_encoding(
        self.get_body_chunk() or b"",
        content_type_header=(
            self.headers.get("Content-Type", None) if self.headers else None
        ),
        markup="xml" if self.document_type == "xml" else "html",
    )"""

    import tempfile

    with tempfile.TemporaryDirectory() as tmpdir:
        # Make a project1/grab/document.py file in the tmpdir with contents of fixture
        Path(tmpdir).joinpath("project1/grab").mkdir(parents=True, exist_ok=True)
        Path(tmpdir).joinpath("project1/grab/document.py").write_text(
            fixture.read_text(),
        )

        problem_w_doc = produce_dy_py_bench_problem(
            task_id=str(1),
            path="grab/document.py",
            identifier="Document.process_encoding",
            func_begin=134,
            func_end=150,
            function=function_w_doc,
            project_id=str(1),
            projects_root=tmpdir,
            strip_docstring=True,
            random_line_seed=3,
            line_to_replace=141,
        )

        yield problem_w_doc, function_w_doc


@pytest.fixture
def example_dypy_prob_fixedline():
    fixture = Path(__file__).parent / "fixtures/document.py.fixture"

    function_w_doc = """def process_encoding(self, encoding: None | str = None) -> str:
    \"\"\"Process explicitly defined encoding or auto-detect it.

    If encoding is explicitly defined, ensure it is a valid encoding the python
    can deal with. If encoding is not specified, auto-detect it.

    Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
    \"\"\"
    if encoding:
        return unicodec.normalize_encoding_name(encoding)
    return unicodec.detect_content_encoding(
        self.get_body_chunk() or b"",
        content_type_header=(
            self.headers.get("Content-Type", None) if self.headers else None
        ),
        markup="xml" if self.document_type == "xml" else "html",
    )"""

    import tempfile

    with tempfile.TemporaryDirectory() as tmpdir:
        # Make a project1/grab/document.py file in the tmpdir with contents of fixture
        Path(tmpdir).joinpath("project1/grab").mkdir(parents=True, exist_ok=True)
        Path(tmpdir).joinpath("project1/grab/document.py").write_text(
            fixture.read_text(),
        )

        problem_w_doc = produce_dy_py_bench_problem(
            task_id=str(1),
            path="grab/document.py",
            identifier="Document.process_encoding",
            func_begin=134,
            func_end=150,
            function=function_w_doc,
            project_id=str(1),
            projects_root=tmpdir,
            strip_docstring=False,
            random_line_seed=3,
            line_to_replace=141,
        )

        yield problem_w_doc, function_w_doc


def test_new_sample_no_eval(example_dypy_prob):
    problem_w_doc, _function_w_doc = example_dypy_prob
    solver = DummyCodeSolverAutoRegressive(None)
    solution = solver.solve(problem_w_doc)
    assert len(solution.solve_steps) == 1
    orig_text = MultilineString(
        problem_w_doc.working_directory.files.get_only_file().content_str,
    )
    text_selector = problem_w_doc.transformation_spec.statements[2].text_selector
    assert isinstance(text_selector, StsSelectorLineSpan)
    verb = problem_w_doc.transformation_spec.statements[2].verb
    assert isinstance(verb, StsEditable)
    assert (
        orig_text[text_selector.start_line_idx]
        == "            return unicodec.normalize_encoding_name(encoding)\n"
    )
    assert (
        solution.solve_steps[0].value
        == "            return unicodec.normalize_encoding_name(encoding)\n"
    )
    solve_text = MultilineString(apply_solution(solution).get_only_file().content_str)
    assert (
        solve_text[134 - 1 : 150]
        == """    def process_encoding(self, encoding: None | str = None) -> str:
        \"\"\"Process explicitly defined encoding or auto-detect it.

        If encoding is explicitly defined, ensure it is a valid encoding the python
        can deal with. If encoding is not specified, auto-detect it.

        Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
        \"\"\"
        if encoding:
            return unicodec.normalize_encoding_name(encoding)
        return unicodec.detect_content_encoding(
            self.get_body_chunk() or b"",
            content_type_header=(
                self.headers.get("Content-Type", None) if self.headers else None
            ),
            markup="xml" if self.document_type == "xml" else "html",
        )\n"""
    )
    assert exact_match_metric(solution).is_success


def test_new_sample_no_eval_nodoc(example_dypy_prob_stripdoc):
    problem_wo_doc, _function_wo_doc = example_dypy_prob_stripdoc
    solver = DummyCodeSolverAutoRegressive(None)
    solution = solver.solve(problem_wo_doc)
    render = render_single_edit_codeproblem_autoregressive(problem_wo_doc)
    assert (
        render.prompt.text
        == "    def process_encoding(self, encoding: None | str = None) -> str:\n      "
        "  if encoding:\n"
    )
    assert len(solution.solve_steps) == 1
    orig_text = MultilineString(
        problem_wo_doc.working_directory.files.get_only_file().content_str,
    )
    text_selector = problem_wo_doc.transformation_spec.statements[-1].text_selector
    assert isinstance(text_selector, StsSelectorLineSpan)
    verb = problem_wo_doc.transformation_spec.statements[-1].verb
    assert isinstance(verb, StsEditable)
    assert (
        orig_text[text_selector.start_line_idx]
        == "            return unicodec.normalize_encoding_name(encoding)\n"
    )
    assert (
        solution.solve_steps[0].value
        == "            return unicodec.normalize_encoding_name(encoding)\n"
    )
    solve_text = MultilineString(apply_solution(solution).get_only_file().content_str)
    assert (
        solve_text[134 - 1 : 150]
        == """    def process_encoding(self, encoding: None | str = None) -> str:
        \"\"\"Process explicitly defined encoding or auto-detect it.

        If encoding is explicitly defined, ensure it is a valid encoding the python
        can deal with. If encoding is not specified, auto-detect it.

        Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
        \"\"\"
        if encoding:
            return unicodec.normalize_encoding_name(encoding)
        return unicodec.detect_content_encoding(
            self.get_body_chunk() or b"",
            content_type_header=(
                self.headers.get("Content-Type", None) if self.headers else None
            ),
            markup="xml" if self.document_type == "xml" else "html",
        )\n"""
    )
    assert exact_match_metric(solution).is_success


def test_new_sample_no_eval_doc_fixed(example_dypy_prob_fixedline):
    problem_doc, _function_doc = example_dypy_prob_fixedline
    solver = DummyCodeSolverAutoRegressive(None)
    solution = solver.solve(problem_doc)
    render = render_single_edit_codeproblem_autoregressive(problem_doc)
    assert "if encoding:" not in render.prompt.text
    assert len(solution.solve_steps) == 1
    orig_text = MultilineString(
        problem_doc.working_directory.files.get_only_file().content_str,
    )
    text_selector = problem_doc.transformation_spec.statements[-1].text_selector
    assert isinstance(text_selector, StsSelectorLineSpan)
    verb = problem_doc.transformation_spec.statements[-1].verb
    assert text_selector.start_line_idx == 141
    assert orig_text[141] == "        if encoding:\n"
    assert isinstance(verb, StsEditable)
    assert orig_text[text_selector.start_line_idx] == "        if encoding:\n"
    assert solution.solve_steps[0].value == "        if encoding:\n"
    solve_text = MultilineString(apply_solution(solution).get_only_file().content_str)
    assert (
        solve_text[134 - 1 : 150]
        == """    def process_encoding(self, encoding: None | str = None) -> str:
        \"\"\"Process explicitly defined encoding or auto-detect it.

        If encoding is explicitly defined, ensure it is a valid encoding the python
        can deal with. If encoding is not specified, auto-detect it.

        Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
        \"\"\"
        if encoding:
            return unicodec.normalize_encoding_name(encoding)
        return unicodec.detect_content_encoding(
            self.get_body_chunk() or b"",
            content_type_header=(
                self.headers.get("Content-Type", None) if self.headers else None
            ),
            markup="xml" if self.document_type == "xml" else "html",
        )\n"""
    )


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No openai key")
def test_new_sample_no_eval_chat(example_dypy_prob_fixedline):
    problem_doc, _function_doc = example_dypy_prob_fixedline
    lm = get_open_ai_lm(OpenAiModelNames.gpt_4o_mini)
    solver = LmCodeSolverAutoRegressive(lm)
    render = solver.prompt_renderer.render(problem_doc, lm)
    print("Prompt")
    print(render.prompt)
    assert render.prompt.is_text_a_chat()
    assert len(render.prompt.text) == 1
    assert render.prompt.text[0].content.startswith("Give a completion to the following code")
    solution = solver.solve(problem_doc)
    print("solution")
    print(solution)
    print("solve step:")
    print(solution.solve_steps[0])
    print("known solution:")
    print(problem_doc.known_solutions[0].solve_steps[0])
    assert solution.solve_steps[0].mark_id == problem_doc.known_solutions[0].solve_steps[0].mark_id
    assert solution.solve_steps[0].path == problem_doc.known_solutions[0].solve_steps[0].path
    val = solution.solve_steps[0].value
    assert len(val.strip().split("\n")) == 1  # just a single line
    assert val.strip() != ""
    applied = apply_solution(solution)
    text = applied.get_only_file().content_str
    # Ok we don't know exactly what the model will give us.
    # so we are just going to check if the output has been integrated
    # into the right part of the file.
    clean_v = val.strip("\n")
    expected_substr = f'''
    def select(self, *args: Any, **kwargs: Any) -> SelectorList[_Element]:
        return XpathSelector(self.tree).select(*args, **kwargs)

    def process_encoding(self, encoding: None | str = None) -> str:
        """Process explicitly defined encoding or auto-detect it.

        If encoding is explicitly defined, ensure it is a valid encoding the python
        can deal with. If encoding is not specified, auto-detect it.

        Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
        """
{clean_v}
            return unicodec.normalize_encoding_name(encoding)
        return unicodec.detect_content_encoding(
'''.strip()
    print("Text")
    print(text)
    print("Expected")
    print(expected_substr)
    assert expected_substr in text


def test_new_sample_no_eval_nodoc_fixed(example_dypy_prob_stripdoc_fixedline):
    problem_wo_doc, _function_wo_doc = example_dypy_prob_stripdoc_fixedline
    solver = DummyCodeSolverAutoRegressive(None)
    solution = solver.solve(problem_wo_doc)
    render = render_single_edit_codeproblem_autoregressive(problem_wo_doc)
    assert (
        render.prompt.text
        == "    def process_encoding(self, encoding: None | str = None) -> str:\n"
    )
    assert len(solution.solve_steps) == 1
    orig_text = MultilineString(
        problem_wo_doc.working_directory.files.get_only_file().content_str,
    )
    text_selector = problem_wo_doc.transformation_spec.statements[-1].text_selector
    assert isinstance(text_selector, StsSelectorLineSpan)
    verb = problem_wo_doc.transformation_spec.statements[-1].verb
    assert text_selector.start_line_idx == 141
    assert orig_text[141] == "        if encoding:\n"
    assert isinstance(verb, StsEditable)
    assert orig_text[text_selector.start_line_idx] == "        if encoding:\n"
    assert solution.solve_steps[0].value == "        if encoding:\n"
    solve_text = MultilineString(apply_solution(solution).get_only_file().content_str)
    assert (
        solve_text[134 - 1 : 150]
        == """    def process_encoding(self, encoding: None | str = None) -> str:
        \"\"\"Process explicitly defined encoding or auto-detect it.

        If encoding is explicitly defined, ensure it is a valid encoding the python
        can deal with. If encoding is not specified, auto-detect it.

        Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
        \"\"\"
        if encoding:
            return unicodec.normalize_encoding_name(encoding)
        return unicodec.detect_content_encoding(
            self.get_body_chunk() or b"",
            content_type_header=(
                self.headers.get("Content-Type", None) if self.headers else None
            ),
            markup="xml" if self.document_type == "xml" else "html",
        )\n"""
    )
    assert exact_match_metric(solution).is_success


def test_new_sample_eval_same_files(example_dypy_prob):
    problem_w_doc, _function_w_doc = example_dypy_prob
    solver = DummyCodeSolverAutoRegressive(None)
    solution = solver.solve(problem_w_doc)
    applied_files = apply_solution(solution)
    assert (
        applied_files.get_only_file().content_str
        == problem_w_doc.working_directory.files.get_only_file().content_str
    )
    assert applied_files == problem_w_doc.working_directory.files


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Dypybench downloads large docker images")
def test_new_sample_eval(example_dypy_prob):
    problem_w_doc, _function_w_doc = example_dypy_prob
    solver = DummyCodeSolverAutoRegressive(None)
    solution = solver.solve(problem_w_doc)
    print(solution)
    print(solution.solve_steps)
    assert apply_solution(solution) == problem_w_doc.working_directory.files
    result = evaluate_code_problem_execution(problem_w_doc, solution)
    assert result.summary.tests > 0
    assert result.summary.errors == 0
    assert result.all_passed()


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Dypybench downloads large docker images")
@pytest.mark.skip(
    reason=(
        "save time on execution. Uncomment if need to debug other more end-to-end tests"
    ),
)
def test_run_no_content(example_dypy_prob):
    problem_w_doc, _ = example_dypy_prob
    assert len(problem_w_doc.test_cases) == 1
    test_case = problem_w_doc.test_cases[0]
    print(test_case)
    assert isinstance(test_case, DiscoveredTestsuite)
    cmd_outputs = run_on_docker(
        docker_context=problem_w_doc.environment.docker_env,
        cmds=test_case.cmds,
    )
    exec_result = test_case.cmds_results_to_suite(cmd_outputs)
    parsed_junit = parse_junit_test_cases(exec_result.xml_result, problem_w_doc)
    assert parsed_junit.all_passed()


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Dypybench downloads large docker images")
@pytest.mark.skip(
    reason=(
        "save time on execution. Uncomment if need to debug other more end-to-end tests"
    ),
)
def test_run_same_content(example_dypy_prob):
    problem_w_doc, _ = example_dypy_prob
    assert len(problem_w_doc.test_cases) == 1
    test_case = problem_w_doc.test_cases[0]
    print(test_case)
    assert isinstance(test_case, DiscoveredTestsuite)
    cmd_outputs = run_on_docker(
        docker_context=problem_w_doc.environment.docker_env,
        cmds=test_case.cmds,
        files=problem_w_doc.working_directory.files,
    )
    exec_result = test_case.cmds_results_to_suite(cmd_outputs)
    parsed_junit = parse_junit_test_cases(exec_result.xml_result, problem_w_doc)
    assert parsed_junit.all_passed()
    assert (
        apply_solution(problem_w_doc.known_solutions[0])
        == problem_w_doc.working_directory.files
    )


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Dypybench downloads large docker images")
def test_new_sample_eval_bad_solve(example_dypy_prob):
    problem_w_doc, _function_w_doc = example_dypy_prob
    prob_files = problem_w_doc.working_directory.files
    assert len(list(prob_files.walk(include_dirs=False))) == 1
    assert "project1/grab/document.py" in prob_files
    solver = DummyCodeSolverAutoRegressive(None)
    solution = solver.solve(problem_w_doc)
    assert len(solution.solve_steps) == 1
    print(solution.solve_steps[0].value)
    orig_text = MultilineString(
        problem_w_doc.working_directory.files.get_only_file().content_str,
    )
    text_selector = problem_w_doc.transformation_spec.statements[2].text_selector
    assert isinstance(text_selector, StsSelectorLineSpan)
    verb = problem_w_doc.transformation_spec.statements[2].verb
    assert isinstance(verb, StsEditable)
    assert (
        orig_text[text_selector.start_line_idx]
        == "            return unicodec.normalize_encoding_name(encoding)\n"
    )
    assert (
        solution.solve_steps[0].value
        == "            return unicodec.normalize_encoding_name(encoding)\n"
    )
    solution.solve_steps = (
        dataclasses.replace(solution.solve_steps[0], value=" " * 12 + "assert False\n"),
    )
    solve_text = MultilineString(apply_solution(solution).get_only_file().content_str)
    assert (
        solve_text[134 - 1 : 150]
        == """    def process_encoding(self, encoding: None | str = None) -> str:
        \"\"\"Process explicitly defined encoding or auto-detect it.

        If encoding is explicitly defined, ensure it is a valid encoding the python
        can deal with. If encoding is not specified, auto-detect it.

        Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
        \"\"\"
        if encoding:
            assert False
        return unicodec.detect_content_encoding(
            self.get_body_chunk() or b"",
            content_type_header=(
                self.headers.get("Content-Type", None) if self.headers else None
            ),
            markup="xml" if self.document_type == "xml" else "html",
        )\n"""
    )

    # Make sure gets on the docker
    r = run_on_docker(
        docker_context=_project_docker_envs[1],
        files=apply_solution(solution),
        cmds=[
            "cat project1/grab/document.py",
        ],
    )
    assert r[-1].completed
    assert r[-1].stdout.decode() == str(solve_text)

    assert is_solution_syntax_likely_valid(solution)
    result = evaluate_code_problem_execution(problem_w_doc, solution)
    assert result.summary.tests > 0
    assert result.summary.failures > 0
    assert not result.all_passed()


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Dypybench downloads large docker images")
def test_new_sample_eval_bad_solve_syntax_errors(example_dypy_prob):
    problem_w_doc, _function_w_doc = example_dypy_prob
    solver = DummyCodeSolverAutoRegressive(None)
    solution = solver.solve(problem_w_doc)
    assert len(solution.solve_steps) == 1
    print(solution.solve_steps[0].value)
    orig_text = MultilineString(
        problem_w_doc.working_directory.files.get_only_file().content_str,
    )
    text_selector = problem_w_doc.transformation_spec.statements[2].text_selector
    assert isinstance(text_selector, StsSelectorLineSpan)
    verb = problem_w_doc.transformation_spec.statements[2].verb
    assert isinstance(verb, StsEditable)
    assert (
        orig_text[text_selector.start_line_idx]
        == "            return unicodec.normalize_encoding_name(encoding)\n"
    )
    assert (
        solution.solve_steps[0].value
        == "            return unicodec.normalize_encoding_name(encoding)\n"
    )
    solution.solve_steps = (
        dataclasses.replace(
            solution.solve_steps[0],
            value=" " * 12 + "assert False + :: sfsd if\n",
        ),
    )
    assert not is_solution_syntax_likely_valid(solution)
    result = evaluate_code_problem_execution(problem_w_doc, solution)
    assert not result.all_passed()


def test_yield_dypy_single():
    try:
        root = _default_projects_root()
    except FileNotFoundError:
        pytest.skip("Dypybench projects root not found")
    problems = yield_dypybench()
    list(problems)
    print("got problem")


# @pytest.mark.skipif(
#    IN_GITHUB_ACTIONS,
# def test_dypybench_pytest_no_steps():
#    ).read_text()
#
#        ),
#        1,
#        ).read_text(),


# @pytest.mark.skipif(
#    IN_GITHUB_ACTIONS,
# def test_dypybench_broken_pytest():
#    assert False
#        1,
#        ).read_text(),


# def test_indent_fix():
#    assert False
#
#    function_w_doc = """def process_encoding(self, encoding: None | str = None) -> str:
#    \"\"\"Process explicitly defined encoding or auto-detect it.
#
#    If encoding is explicitly defined, ensure it is a valid encoding the python
#    can deal with. If encoding is not specified, auto-detect it.
#
#    Raises unicodec.InvalidEncodingName if explicitly set encoding is invalid.
#    \"\"\"
#    if encoding:
#    return unicodec.detect_content_encoding(
#        ),
#    )"""
#
#
#        problem_w_doc.format_with_answer(
#            indented_answer,
#        ),
#    assert (
#
#    assert (
#        indented_answer
#        == "            return unicodec.normalize_encoding_name(encoding)"
#
#        "\t",
#
#        "\t",
#
#        "\t",
#
#        "\t",
#



from collections.abc import Sequence

from synthegrator.code_problems import DiscoveredTestsuite
from synthegrator.sandboxing import CmdExecResult, TestsuiteExecResult


def custom_suite_extractor(cmds: Sequence[CmdExecResult]) -> TestsuiteExecResult:
    return TestsuiteExecResult()


def test_discovered_suite_hashing():
    tc = DiscoveredTestsuite(["pytest", "cat result"])
    assert tc.get_hash() == tc.get_hash()
    print(tc.get_hash())
    # NOTE: This is just verifying that it's stable between runs.
    #     Going to be a bit of pain to maintain (have to just copy new value
    #     anytime the default extractor changes). If it becomes a problem can
    #     just remove this part.
    assert tc.get_hash() == "9390d2b96f83cb97"
    tc2 = DiscoveredTestsuite(["pytest", "cat not"])
    assert tc.get_hash() != tc2.get_hash()

    tc3 = DiscoveredTestsuite(
        ["pytest", "cat not"],
        cmds_results_to_suite=custom_suite_extractor,
    )

    assert tc3.get_hash() != tc2.get_hash()



import dataclasses
import difflib
from itertools import islice
from pprint import pprint
from unittest.mock import MagicMock

import pytest

from synthegrator.code_problems import CodeProblem
from synthegrator.code_solver import DummyCodeSolverAutoRegressive
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.problem_rendering import (
    PromptRendererSingleEditGeneric,
    render_single_edit_codeproblem_autoregressive,
)
from synthegrator.solution_eval import apply_solution, evaluate_code_problem_execution
from synthegrator.synthdatasets.mbpp import (
    MbppSplit,
    _get_mbpp_huggingface,
    _get_mbpp_plus_huggingface,
    _iterate_for_split,
    yield_mbpp,
    yield_mbpp_plus,
)
from synthegrator.transformation_spec import SolveStep
from synthegrator.util import pretty_print_python_code


def test_get_single_instance():
    val = next(yield_mbpp())
    assert val.problem_id == "601"

    assert (
        val.instructions
        == "You are an expert Python programmer, and here is your task: Write a"
        " function to find the longest chain which can be formed from the given set"
        " of pairs. Your code should pass these tests:\n\nassert"
        " max_chain_length([Pair(5, 24), Pair(15, 25),Pair(27, 40), Pair(50, 60)],"
        " 4) == 3\nassert max_chain_length([Pair(1, 2), Pair(3, 4),Pair(5, 6),"
        " Pair(7, 8)], 4) == 4\nassert max_chain_length([Pair(19, 10), Pair(11,"
        " 12),Pair(13, 14), Pair(15, 16), Pair(31, 54)], 5) == 5\n\n"
    )
    assert val.dataset_name == "mbpp"


def test_get_single_instance_method_completion():
    val = next(yield_mbpp(format_as_method_completion_problem=True))
    assert val.problem_id == "601func"
    assert val.dataset_name == "mbpp"
    render = render_single_edit_codeproblem_autoregressive(val)
    pretty_print_python_code(render.prompt.text)
    # fmt: off
    assert render.prompt.text == (
        'class Pair(object): \n'
        '    def __init__(self, a, b): \n'
        '        self.a = a \n'
        '        self.b = b \n'
        'def max_chain_length(arr, n): \n'
        '    """\n'
        '    Write a function to find the longest chain which can be formed from the given set of pairs.\n'
        '    \n'
        '    Examples:\n'
        '    assert max_chain_length([Pair(5, 24), Pair(15, 25),Pair(27, 40), Pair(50, 60)], 4) == 3\n'
        '    assert max_chain_length([Pair(1, 2), Pair(3, 4),Pair(5, 6), Pair(7, 8)], 4) == 4\n'
        '    assert max_chain_length([Pair(19, 10), Pair(11, 12),Pair(13, 14), Pair(15, 16), Pair(31, 54)], 5) == 5\n'
        '    """\n'
    )
    sol = apply_solution(val.known_solutions[0]).get_only_file().content_str
    assert sol == (
        'class Pair(object): \n'
        '    def __init__(self, a, b): \n'
        '        self.a = a \n'
        '        self.b = b \n'
        'def max_chain_length(arr, n): \n'
        '    """\n'
        '    Write a function to find the longest chain which can be formed from the given set of pairs.\n'
        '    \n'
        '    Examples:\n'
        '    assert max_chain_length([Pair(5, 24), Pair(15, 25),Pair(27, 40), Pair(50, 60)], 4) == 3\n'
        '    assert max_chain_length([Pair(1, 2), Pair(3, 4),Pair(5, 6), Pair(7, 8)], 4) == 4\n'
        '    assert max_chain_length([Pair(19, 10), Pair(11, 12),Pair(13, 14), Pair(15, 16), Pair(31, 54)], 5) == 5\n'
        '    """\n'
        '    max = 0\n'
        '    mcl = [1 for i in range(n)] \n'
        '    for i in range(1, n): \n'
        '        for j in range(0, i): \n'
        '            if (arr[i].a > arr[j].b and\n'
        '                mcl[i] < mcl[j] + 1): \n'
        '                mcl[i] = mcl[j] + 1\n'
        '    for i in range(n): \n'
        '        if (max < mcl[i]): \n'
        '            max = mcl[i] \n'
        '    return max\n'
    )
    # fmt: on


@pytest.mark.parametrize(
    "format_as_method_completion_problem",
    [False, True],
    ids=["instruction", "method_completion"],
)
def test_get_all(format_as_method_completion_problem):
    assert len(list(yield_mbpp(MbppSplit.TRAIN))) == 374
    assert len(list(yield_mbpp(MbppSplit.VALIDATION))) == 90
    assert len(list(yield_mbpp())) == 974
    assert len(list(yield_mbpp(sanitized_only=True))) == 427


def test_apply_prime_num():
    for prob in islice(
        yield_mbpp(
            MbppSplit.ALL,
            only_with_passing_reference_solution=True,
            format_as_method_completion_problem=True,
        ),
        3,
        4,
    ):
        assert len(prob.known_solutions) == 1
        assert prob.problem_id == "605func"  # The method prime_num problem which has issues
        apply_solution(prob.known_solutions[0])


# def test_apply_prob911():
#    # This problem has a `assert int(fuc())` style
#    for prob in islice(yield_mbpp(
#        MbppSplit.ALL,
#    ), 911 - 601, 911 - 601 + 1):


@pytest.mark.parametrize(
    "format_as_method_completion_problem",
    [False, True],
    ids=["instruction", "method_completion"],
)
def test_apply_all_solutions(format_as_method_completion_problem):
    for prob in yield_mbpp(
        MbppSplit.ALL,
        only_with_passing_reference_solution=True,
        format_as_method_completion_problem=format_as_method_completion_problem,
    ):
        assert len(prob.known_solutions) == 1
        print(prob.problem_id)
        apply_solution(prob.known_solutions[0])


def test_whitespace_on_prob():
    for prob in yield_mbpp(format_as_method_completion_problem=True):
        if prob.problem_id == "226func":
            break
    else:
        msg = "Problem not found"
        raise AssertionError(msg)
    renderer = PromptRendererSingleEditGeneric()
    lm = MagicMock()
    lm.is_chat_model = False
    render = renderer.render(prob, lm)
    # fmt: off
    assert render.prompt.text == (
        '''def odd_values_string(str):
    """
    Write a python function to remove the characters which have odd index values of a given string.
    
    Examples:
    assert odd_values_string('abcdef') == 'ace'
    assert odd_values_string('python') == 'pto'
    assert odd_values_string('data') == 'dt'
    assert odd_values_string('lambs') == 'lms'
    """
'''
    )
    # fmt: on


@pytest.mark.parametrize(
    "format_as_method_completion_problem",
    [False, True],
    ids=["instruction", "method_completion"],
)
def test_dummy_solve(format_as_method_completion_problem):
    dummy_solve = DummyCodeSolverAutoRegressive(MagicMock())
    failing_probs = []
    real_exs = _get_mbpp_huggingface(sanitized=False)
    real_exs = {
        str(ex["task_id"]): ex for ex in _iterate_for_split(real_exs, MbppSplit.ALL)
    }
    for prob in islice(
        yield_mbpp(
            MbppSplit.ALL,
            only_with_passing_reference_solution=True,
            format_as_method_completion_problem=format_as_method_completion_problem,
        ),
        None,
        None,
        100,
    ):
        solution = dummy_solve.solve(prob, PythonLangSpec())
        result = evaluate_code_problem_execution(prob, solution)
        if not result.all_passed():
            print("problem id")
            print(prob.problem_id)
            print("--- Application:")
            pretty_print_python_code(
                apply_solution(solution).get_only_file().content_str,
            )
            print(result)
            print(prob)
            print("Tc")
            print(result)
            print("Tc")
            for tc in real_exs[prob.problem_id]["test_list"]:
                pretty_print_python_code(tc)
            print("Instructions:")
            print(prob.instructions)
            print("-- Real")
            pretty_print_python_code(real_exs[prob.problem_id]["code"])
            print("-- Render:")
            pretty_print_python_code(
                render_single_edit_codeproblem_autoregressive(prob).text,
            )
            failing_probs.append(prob)
            raise AssertionError
    print(failing_probs)
    print("failing ids", [p.problem_id for p in failing_probs])
    assert len(failing_probs) == 0


@pytest.mark.parametrize(
    "format_as_method_completion_problem",
    # [False, True],
    # ids=["instruction", "method_completion"],
    [False],
    ids=["instruction"],
)
def test_mbpp_plus(format_as_method_completion_problem):
    dummy_solve = DummyCodeSolverAutoRegressive(MagicMock())
    failing_probs = []
    real_exs = _get_mbpp_plus_huggingface()
    real_exs = {str(ex["task_id"]): ex for ex in real_exs["test"]}
    probs = list(
        yield_mbpp_plus(
            format_as_method_completion_problem=format_as_method_completion_problem,
        )
    )
    assert len(probs) == 378
    for prob in islice(
        probs,
        None,
        None,
        100,
    ):
        solution = dummy_solve.solve(prob, PythonLangSpec())
        result = evaluate_code_problem_execution(prob, solution)
        if not result.all_passed():
            print("problem id")
            print(prob.problem_id)
            print("--- Application:")
            pretty_print_python_code(
                apply_solution(solution).get_only_file().content_str,
            )
            print(result)
            print(prob)
            print("Tc")
            print(result)
            print("Tc")
            for tc in real_exs[prob.problem_id]["test_list"]:
                pretty_print_python_code(tc)
            print("Instructions:")
            print(prob.instructions)
            print("-- Real")
            pretty_print_python_code(real_exs[prob.problem_id]["code"])
            print("-- Render:")
            pretty_print_python_code(
                render_single_edit_codeproblem_autoregressive(prob).text,
            )
            failing_probs.append(prob)
            raise AssertionError
    print(failing_probs)
    print("failing ids", [p.problem_id for p in failing_probs])
    assert len(failing_probs) == 0


def test_a_solve_that_passes_mbpp_but_not_plus():
    first_plus_prob = next(
        yield_mbpp_plus(
            format_as_method_completion_problem=True,
        )
    )
    probs = list(
        yield_mbpp(
            split=MbppSplit.ALL,
            format_as_method_completion_problem=True,
        )
    )
    mbpp_instruct_to_p: dict[str, CodeProblem] = {
        p.working_directory.files.get_only_file().content_str: p for p in probs
    }
    assert len(probs) == len(mbpp_instruct_to_p)
    pprint("all problems")
    print("problem id", first_plus_prob.instructions)
    content = first_plus_prob.working_directory.files.get_only_file().content_str
    print("Content")
    print("{" + content + "}")
    for key, prob in mbpp_instruct_to_p.items():
        if "is_not_prime" in key:
            print("FOUND")
            print("{" + key + "}")
            for tc in prob.test_cases:
                print("TC", tc.method_body)
            """
            TC assert is_not_prime(2) == False
            TC assert is_not_prime(10) == True
            TC assert is_not_prime(35) == True
            TC assert is_not_prime(37) == False
            """
            try:
                assert key == content
            except AssertionError:
                diff = difflib.unified_diff(
                    key.splitlines(keepends=True),
                    content.splitlines(keepends=True),
                    fromfile="key",
                    tofile="content",
                    lineterm="",
                )
                diff_text = "".join(diff)
                stripped_match = key.strip() == content.strip()
                raise AssertionError(
                    f"Strings are not equal. Stripped match: {stripped_match}. Diff:\n{diff_text}"
                ) from None
    normal_version = mbpp_instruct_to_p[content]
    print(normal_version)
    print("normal version")
    print(normal_version.instructions)
    print(print(normal_version))
    # fmt: off
    stupid_version = (
"""
    return {
        2: False,
        10: True,
        35: True,
        37: False,
    }.get(n, True)
"""
    )
    # fmt: on
    normal_sol = normal_version.known_solutions[0]
    stupid_solve = dataclasses.replace(
        normal_sol,
        solve_steps=[
            SolveStep(
                path=normal_sol.solve_steps[0].path,
                mark_id=normal_sol.solve_steps[0].mark_id,
                value=stupid_version,
            )
        ],
    )
    stupid_eval_normal = evaluate_code_problem_execution(normal_version, stupid_solve)
    assert stupid_eval_normal.all_passed()
    new_solve = normal_version.known_solutions[0]
    stupid_solve = dataclasses.replace(
        new_solve,
        solve_steps=[
            SolveStep(
                path=new_solve.solve_steps[0].path,
                mark_id=new_solve.solve_steps[0].mark_id,
                value=stupid_version,
            )
        ],
    )
    stupid_eval_new = evaluate_code_problem_execution(first_plus_prob, stupid_solve)
    assert not stupid_eval_new.all_passed()



from synthegrator.code_solver import LmCodeSolverAutoRegressive
import json
from pathlib import Path
from lmwrapper.openai_wrapper import get_open_ai_lm, OpenAiModelNames
from synthegrator.execution_threading import solve_and_evaluate_problems
from synthegrator.solution_eval import evaluate_code_problem_execution
from synthegrator.synthdatasets.repocod import yield_repocod_problems


cur_path = Path(__file__).parent


def _look_more_at_df():
    from ._quick_inspect_df import _make_df
    from datasets import load_dataset
    from lmwrapper.openai_wrapper import get_open_ai_lm, OpenAiModelNames
    from ._make_author_style import calc_results_for_author_style
    df, key = _make_df(1000, OpenAiModelNames.gpt_4o_mini, "autoregressive")
    print("is success")
    print(df.main_metric__is_success.mean())
    print("collection error")
    print(df.test_results_list__collection_error.mean())
    print("Execution error")
    print(df.test_results_list__exec_error.mean())
    print("Timeout")
    print(df.test_results_list__timeout.mean())
    print("test_results_list__len distribution")
    print(df.test_results_list__len.describe())
    print("Frac len < 1")
    print((df.test_results_list__len < 1).mean())
    author_run_df = calc_results_for_author_style("mini_synth_980")
    
    # Add author_ prefix to all columns except problem_id
    author_run_df_prefixed = author_run_df.copy()
    columns_to_rename = [col for col in author_run_df_prefixed.columns if col != 'problem_id']
    author_run_df_prefixed = author_run_df_prefixed.rename(columns={col: f'author_{col}' for col in columns_to_rename})
    
    # Join the dataframes on problem_id
    joined_df = df.merge(
        author_run_df_prefixed, 
        left_on='solution__problem__problem_id', 
        right_on='problem_id', 
        how='inner'
    )
    assert len(joined_df) == len(df)
    
    print(joined_df.columns)
    # Compare when author_result differs from main_metric__is_success
    
    # Basic difference statistics
    print("\n=== BASIC COMPARISON STATS ===")
    print(f"Total problems: {len(joined_df)}")
    print(f"Problems with author data: {joined_df['author_result'].notna().sum()}")
    
    # Filter to only rows where we have both results
    comparison_df = joined_df[joined_df['author_result'].notna()].copy()
    print(f"Problems with both results: {len(comparison_df)}")
    
    # Basic success rates
    main_success_rate = comparison_df['main_metric__is_success'].mean()
    author_success_rate = comparison_df['author_result'].mean()
    print(f"Main model success rate: {main_success_rate:.3f}")
    print(f"Author model success rate: {author_success_rate:.3f}")
    print(f"Difference: {author_success_rate - main_success_rate:.3f}")
    
    # Agreement analysis
    agreement = (comparison_df['main_metric__is_success'] == comparison_df['author_result']).sum()
    disagreement = len(comparison_df) - agreement
    print(f"\nAgreement: {agreement} ({agreement/len(comparison_df):.3f})")
    print(f"Disagreement: {disagreement} ({disagreement/len(comparison_df):.3f})")
    
    # Break down disagreements
    main_pass_author_fail = ((comparison_df['main_metric__is_success'] == True) & 
                            (comparison_df['author_result'] == False)).sum()
    main_fail_author_pass = ((comparison_df['main_metric__is_success'] == False) & 
                            (comparison_df['author_result'] == True)).sum()
    
    print(f"\nMain PASS, Author FAIL: {main_pass_author_fail}")
    print(f"Main FAIL, Author PASS: {main_fail_author_pass}")
    
    # Look at other metrics for disagreement cases
    print("\n=== DESCRIPTIVE ANALYSIS FOR DISAGREEMENTS ===")
    
    # Cases where main passes but author fails
    if main_pass_author_fail > 0:
        main_pass_author_fail_df = comparison_df[
            (comparison_df['main_metric__is_success'] == True) & 
            (comparison_df['author_result'] == False)
        ]
        print(f"\nMain PASS, Author FAIL cases ({main_pass_author_fail}):")
        print("Test metrics for these cases:")
        print(f"  Avg collection error: {main_pass_author_fail_df['test_results_list__collection_error'].mean():.3f}")
        print(f"  Avg execution error: {main_pass_author_fail_df['test_results_list__exec_error'].mean():.3f}")
        print(f"  Avg timeout: {main_pass_author_fail_df['test_results_list__timeout'].mean():.3f}")
        print(f"  Avg test length: {main_pass_author_fail_df['test_results_list__len'].mean():.3f}")
        
        # Show some example problem IDs
        print("  Example problem IDs:")
        for pid in main_pass_author_fail_df['solution__problem__problem_id'].head(5):
            print(f"    {pid}")
    
    # Cases where main fails but author passes  
    if main_fail_author_pass > 0:
        main_fail_author_pass_df = comparison_df[
            (comparison_df['main_metric__is_success'] == False) & 
            (comparison_df['author_result'] == True)
        ]
        print(f"\nMain FAIL, Author PASS cases ({main_fail_author_pass}):")
        print("Test metrics for these cases:")
        print(f"  Avg collection error: {main_fail_author_pass_df['test_results_list__collection_error'].mean():.3f}")
        print(f"  Avg execution error: {main_fail_author_pass_df['test_results_list__exec_error'].mean():.3f}")
        print(f"  Avg timeout: {main_fail_author_pass_df['test_results_list__timeout'].mean():.3f}")
        print(f"  Avg test length: {main_fail_author_pass_df['test_results_list__len'].mean():.3f}")
        
        # Show some example problem IDs
        print("  Example problem IDs:")
        for pid in main_fail_author_pass_df['solution__problem__problem_id'].head(5):
            print(f"    {pid}")
    
    # Repository-level analysis
    print("\n=== REPOSITORY-LEVEL ANALYSIS ===")
    if 'author_repo_name' in comparison_df.columns:
        # Success rates by repository
        repo_comparison = comparison_df.groupby('author_repo_name').agg({
            'main_metric__is_success': 'mean',
            'author_result': 'mean'
        }).round(3)
        repo_comparison['difference'] = repo_comparison['author_result'] - repo_comparison['main_metric__is_success']
        repo_comparison = repo_comparison.sort_values('difference', ascending=False)
        print("Success rates by repository:")
        print(repo_comparison)
        
        # Count-based analysis by repository
        print("\n=== COUNT-BASED ANALYSIS BY REPOSITORY ===")
        repo_counts = comparison_df.groupby('author_repo_name').agg({
            'main_metric__is_success': ['sum', 'count'],
            'author_result': 'sum'
        })
        
        # Flatten column names
        repo_counts.columns = ['main_successes', 'total_problems', 'author_successes']
        repo_counts['main_failures'] = repo_counts['total_problems'] - repo_counts['main_successes']
        repo_counts['author_failures'] = repo_counts['total_problems'] - repo_counts['author_successes']
        repo_counts['success_diff'] = repo_counts['author_successes'] - repo_counts['main_successes']
        
        # Sort by absolute difference to see biggest impact
        repo_counts = repo_counts.sort_values('success_diff', ascending=False)
        print("Count differences by repository (author_successes - main_successes):")
        print(repo_counts[['total_problems', 'main_successes', 'author_successes', 'success_diff']])
        
        # Show repositories with biggest positive and negative differences
        print(f"\nRepositories where author approach helps most:")
        positive_diff = repo_counts[repo_counts['success_diff'] > 0]
        if len(positive_diff) > 0:
            for repo in positive_diff.head(3).index:
                diff = positive_diff.loc[repo, 'success_diff']
                total = positive_diff.loc[repo, 'total_problems']
                print(f"  {repo}: +{diff} successes out of {total} problems")
        
        print(f"\nRepositories where author approach hurts most:")
        negative_diff = repo_counts[repo_counts['success_diff'] < 0]
        if len(negative_diff) > 0:
            for repo in negative_diff.tail(3).index:
                diff = negative_diff.loc[repo, 'success_diff']
                total = negative_diff.loc[repo, 'total_problems']
                print(f"  {repo}: {diff} successes out of {total} problems")
    
    # Deep dive into scikit-learn since it's the biggest issue
    print("\n=== SCIKIT-LEARN DEEP DIVE ===")
    sklearn_df = comparison_df[comparison_df['author_repo_name'] == 'scikit-learn'].copy()
    if len(sklearn_df) > 0:
        print(f"Total scikit-learn problems: {len(sklearn_df)}")
        print(f"Main model successes: {sklearn_df['main_metric__is_success'].sum()}")
        print(f"Author model successes: {sklearn_df['author_result'].sum()}")
        print(f"Success rate drop: {sklearn_df['main_metric__is_success'].mean():.3f} → {sklearn_df['author_result'].mean():.3f}")
        
        # Look at test execution metrics for scikit-learn
        print("\nTest execution metrics for scikit-learn:")
        print(f"  Collection error rate: {sklearn_df['test_results_list__collection_error'].mean():.3f}")
        print(f"  Execution error rate: {sklearn_df['test_results_list__exec_error'].mean():.3f}")
        print(f"  Timeout rate: {sklearn_df['test_results_list__timeout'].mean():.3f}")
        print(f"  Syntax check success rate: {sklearn_df['test_results_list__syntax_check_result'].mean():.3f}")
        print(f"  Average test length: {sklearn_df['test_results_list__len'].mean():.1f}")
        print(f"  Average runtime: {sklearn_df['test_results_list__runtime'].mean():.3f}")
        
        # Compare main successes vs author successes within scikit-learn
        sklearn_main_success = sklearn_df[sklearn_df['main_metric__is_success'] == True]
        sklearn_author_success = sklearn_df[sklearn_df['author_result'] == True]
        
        print(f"\nFor main model successes in scikit-learn ({len(sklearn_main_success)}):")
        print(f"  Avg collection error: {sklearn_main_success['test_results_list__collection_error'].mean():.3f}")
        print(f"  Avg execution error: {sklearn_main_success['test_results_list__exec_error'].mean():.3f}")
        print(f"  Avg timeout: {sklearn_main_success['test_results_list__timeout'].mean():.3f}")
        print(f"  Avg test length: {sklearn_main_success['test_results_list__len'].mean():.1f}")
        
        print(f"\nFor author model successes in scikit-learn ({len(sklearn_author_success)}):")
        print(f"  Avg collection error: {sklearn_author_success['test_results_list__collection_error'].mean():.3f}")
        print(f"  Avg execution error: {sklearn_author_success['test_results_list__exec_error'].mean():.3f}")
        print(f"  Avg timeout: {sklearn_author_success['test_results_list__timeout'].mean():.3f}")
        print(f"  Avg test length: {sklearn_author_success['test_results_list__len'].mean():.1f}")
        
        # Look at the overlap - problems that both succeed or both fail
        both_succeed = sklearn_df[(sklearn_df['main_metric__is_success'] == True) & 
                                 (sklearn_df['author_result'] == True)]
        both_fail = sklearn_df[(sklearn_df['main_metric__is_success'] == False) & 
                              (sklearn_df['author_result'] == False)]
        main_only = sklearn_df[(sklearn_df['main_metric__is_success'] == True) & 
                              (sklearn_df['author_result'] == False)]
        author_only = sklearn_df[(sklearn_df['main_metric__is_success'] == False) & 
                                (sklearn_df['author_result'] == True)]
        
        print(f"\nOverlap analysis for scikit-learn:")
        print(f"  Both succeed: {len(both_succeed)}")
        print(f"  Both fail: {len(both_fail)}")
        print(f"  Main only succeeds: {len(main_only)}")
        print(f"  Author only succeeds: {len(author_only)}")
        
        # Show some examples of main-only successes
        if len(main_only) > 0:
            print(f"\nExample problems where main succeeds but author fails:")
            for i, (idx, row) in enumerate(main_only.head(5).iterrows()):
                print(f"  {i+1}. {row['solution__problem__problem_id']}")
                print(f"     Collection error: {row['test_results_list__collection_error']}")
                print(f"     Execution error: {row['test_results_list__exec_error']}")
                print(f"     Test length: {row['test_results_list__len']}")
                print(f"     repo_id: {row['author_repo_id']}")
                print(row)
                exit()
    
    return comparison_df


def _resolve_some():
    probs = []
    for p in yield_repocod_problems():
        if "Repository-level_ARDRegression.fit" in p.problem_id:
            probs.append(p)
            break

    solver = LmCodeSolverAutoRegressive(
        get_open_ai_lm(OpenAiModelNames.gpt_4o_mini),
        include_lm_response=True,
    )
    evals = solve_and_evaluate_problems(solver, probs)
    author_test_results = (cur_path / "authorcode/evaluate/pytest_results/mini_synth_980/scikit-learn/function_0/pytest_result.json").read_text()
    author_test_results = json.loads(author_test_results)
    for ev in evals:
        print(ev)
        print(dir(ev.solution))
        #print(ev.solution.lm_prediction.prompt.temperature)
        print(ev.solution.solve_steps[0].value)
        print(ev.test_results)
        print(author_test_results.keys())
        print(author_test_results['summary'])
        for test in ev.test_results:
            print(test)
        for path, f in ev.solution.apply().walk(only_consider_dirty=True):
            #print(f.content_str)
            (cur_path / "export_save").write_text(f.content_str)
            print(path)
        print(ev.solution.problem.environment.docker_env)
        nev = evaluate_code_problem_execution(
            ev.solution.problem, 
            ev.solution,
            try_cache=False,
            do_not_execute_if_syntax_fail=False,
        )
        print(nev)
        exit()


if __name__ == "__main__":
    #_look_more_at_df()
    _resolve_some()


from synthegrator.lang_specs.lang_spec import PythonFunctionParser
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec


def test_get_function():
    src = "def foo():\n    pass\n"
    funcs = list(PythonLangSpec().find_functions(src))
    assert len(funcs) == 1
    func = funcs[0]
    assert func.get_function_name() == "foo"
    assert func.get_declaration_str() == "def foo():\n"
    assert func.get_declaration_str(include_body_indent=True) == "def foo():\n    "
    assert func.get_body_src() == "    pass\n"
    assert func.get_body_src(include_body_indent=False) == "pass\n"
    assert func.get_prefix_comment_str() is None


def test_get_function_multiline():
    src = "def foo():\n    x = 2\n    return x * 2\n"
    funcs = list(PythonLangSpec().find_functions(src))
    assert len(funcs) == 1
    func = funcs[0]
    assert func.get_function_name() == "foo"
    assert func.get_declaration_str() == "def foo():\n"
    assert func.get_body_src() == "    x = 2\n    return x * 2\n"
    assert func.get_prefix_comment_str() is None


def test_get_function_args():
    src = "def foo(a):\n    pass\n"
    funcs = list(PythonLangSpec().find_functions(src))
    assert len(funcs) == 1
    func = funcs[0]
    assert func.get_function_name() == "foo"
    assert func.get_declaration_str() == "def foo(a):\n"
    assert func.get_body_src() == "    pass\n"
    assert func.get_prefix_comment_str() is None


def test_get_function_args2():
    src = "def foo(a, b):\n    pass\n"
    funcs = list(PythonLangSpec().find_functions(src))
    assert len(funcs) == 1
    func = funcs[0]
    assert func.get_function_name() == "foo"
    assert func.get_declaration_str() == "def foo(a, b):\n"
    assert func.get_body_src() == "    pass\n"
    assert func.get_prefix_comment_str() is None


def test_get_function_args_type_hint():
    src = "def foo(a: int, b: list[tuple[str, ...]]):\n    pass\n"
    funcs = list(PythonLangSpec().find_functions(src))
    assert len(funcs) == 1
    func = funcs[0]
    assert func.get_function_name() == "foo"
    assert func.get_declaration_str() == src.splitlines(keepends=True)[0]
    assert func.get_body_src() == "    pass\n"
    assert func.get_prefix_comment_str() is None


def test_get_two_funcs():
    src = "def foo():\n    pass\n\ndef bar():\n    pass\n"
    funcs = list(PythonLangSpec().find_functions(src))
    assert len(funcs) == 2
    func = funcs[0]
    assert func.get_function_name() == "foo"
    assert func.get_declaration_str() == src.splitlines(keepends=True)[0]
    assert func.get_body_src() == "    pass\n"
    assert func.get_prefix_comment_str() is None
    func = funcs[1]
    assert func.get_function_name() == "bar"
    assert func.get_declaration_str() == src.splitlines(keepends=True)[3]
    assert func.get_body_src() == "    pass\n"
    assert func.get_prefix_comment_str() is None


def test_get_function_multiline_decl():
    src = "def foo(\n   a: int, \n   b: list[tuple[str, ...]]\n) -> str:\n    pass\n"
    funcs = list(PythonLangSpec().find_functions(src))
    assert len(funcs) == 1
    func = funcs[0]
    assert func.get_function_name() == "foo"
    assert func.get_declaration_str() == src[: src.index("pass") - 4]
    assert func.get_body_src() == "    pass\n"
    assert func.get_prefix_comment_str() is None


def test_line_numbers():
    doc = """
# hi

def h(x,
    y,
    z,
    k) -> None:
    \"\"\"hi\"\"\"
    logging.info(x, y)
    return None"""

    doc_f = PythonFunctionParser(doc)
    assert doc_f.body_span_begin == 8
    assert doc_f.body_span_end == 9

    short_function = PythonFunctionParser(
        """def hi():
    pass""",
    )
    assert short_function.body_span_begin == 1


def test_funtion_parser():
    definition = """def h(x,
    y) -> None:"""

    def_w_doc = """def h(x,
    y) -> None:
    \"\"\"hi\"\"\""""

    doc = """def h(x,
    y) -> None:
    \"\"\"hi\"\"\"
    logging.info(x, y)"""

    no_doc = """def h(x,
    y) -> None:
    logging.info(x, y)"""

    doc_f = PythonFunctionParser(doc)
    assert doc_f.has_docstring is True
    assert doc_f.function_without_doc == no_doc
    assert doc_f.function_definition_with_doc == def_w_doc
    assert doc_f.function_definition_without_doc == definition
    assert doc_f.body_span_begin == 3

    doc_nf = PythonFunctionParser(no_doc)
    assert doc_nf.has_docstring is False
    assert doc_nf.function_without_doc == no_doc
    assert doc_nf.function_definition_with_doc == definition
    assert doc_nf.function_definition_without_doc == definition



from abc import ABC, abstractmethod
from collections.abc import Sequence

from lmwrapper.structs import LmPrediction

from synthegrator.code_problems import CodeProblem
from synthegrator.lang_specs.lang_spec import LangSpec
from synthegrator.problem_rendering import LmPromptRender, LmPromptRenderSingleEdit
from synthegrator.transformation_spec import (
    MarkElement,
    SolveStep,
    StsEditable,
    StsInsert,
    StsPlaceTransforming,
)
from synthegrator.util import get_first_line_with_nonwhitespace


def format_return_val_for_node(
    value: str,
    node: MarkElement[StsPlaceTransforming],
    lang_spec,
) -> str | tuple[str, tuple[int, int]]:
    """
    Makes sure that a solution value to be put in place of the node
    matches all the node's requirements (like starting with certain indentation
    or whatever).

    If you give it joinable_tokens then will also return a tuple
    representing the best guess of the at the slice of the tokens that
    are actually used in this new value. Note that will not always be an exact
    match. For exampling setting the base indent or setting the ending text
    will not have any analog to the original tokens.
    """
    verb = node.verb
    if not isinstance(verb, StsPlaceTransforming):
        msg = "Verb must be StsPlaceTransforming"
        raise TypeError(msg)
    if not isinstance(verb, StsInsert | StsEditable):
        msg = f"Unhandled verb type: {type(verb)}"
        raise TypeError(msg)

    if verb.set_base_space_indent:
        value = set_base_indent(value.lstrip("\n"), verb.set_base_space_indent)
    if verb.stop_at_block_end:
        value = lang_spec.truncate_function_completion(value)
    if verb.set_ending_text:
        new_start = _get_new_content_start(value, verb.set_ending_text)
        value += verb.set_ending_text[new_start:]
    if verb.max_length_chars is not None:
        raise NotImplementedError
    if verb.max_length_lines is not None:
        lines = value.splitlines(keepends=True)
        value = "".join(lines[: min(verb.max_length_lines, len(lines))])

    return value


class AnswerForSolutionNeverSupported(NotImplementedError):
    """
    Raised when try to get an answer for a solution,
    but it is not implemented for the response parser.
    """


class AnswerForSolutionInvalidForSolve(ValueError):
    """
    Raised when try to get an answer for a solution,
    but there is no known way to construct an answer
    that would produce the solve steps given.
    """


class ResponseParser(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def parse(
        self,
        render: LmPromptRender,
        resp: LmPrediction,
        problem: CodeProblem,
    ) -> list[SolveStep]:
        raise NotImplementedError

    # TODO: consider if can do this more naturally with Protocol/mixin
    def answer_for_solution(
        self,
        render: LmPromptRender,
        problem: CodeProblem,
        solve_steps: Sequence[SolveStep],
    ) -> str:
        raise AnswerForSolutionNeverSupported


class ResponseParserSingleEdit(ResponseParser):
    def __init__(self):
        super().__init__()

    def parse(
        self,
        render: LmPromptRenderSingleEdit,
        resp: LmPrediction,
        problem: CodeProblem,
    ) -> list[SolveStep]:
        lang_spec = problem.get_lang_spec_for_path(render.path)
        if problem.instructions_are_essential or render.prompt.is_text_a_chat():
            solve_steps = pull_solution_from_dialog_completion(
                resp,
                problem,
                lang_spec,
                code_prefix=render.code_prefix,
                insert_node=render.changed_element,
                path=render.path,
            )
        else:
            solution_text = format_return_val_for_node(
                resp.completion_text,
                render.changed_element,
                lang_spec,
            )
            solve_steps = (
                SolveStep(
                    render.path,
                    render.changed_element.mark_id,
                    solution_text,
                ),
            )
        return solve_steps


def set_base_indent(code_text: str, space_indent_level: int) -> str:
    lines = code_text.split("\n")
    first_line = get_first_line_with_nonwhitespace(code_text)
    current_base_indent = len(first_line) - len(first_line.lstrip())
    delta = space_indent_level - current_base_indent

    if delta == 0:
        return code_text

    if delta > 0:
        return "\n".join(
            [" " * delta + line if line.strip() else line for line in lines],
        )

    # delta < 0
    return "\n".join([line[-delta:] for line in lines])


def pull_solution_from_dialog_completion(
    completion: LmPrediction,
    problem: CodeProblem,
    lang_spec: LangSpec,
    code_prefix: str,
    insert_node: MarkElement[StsPlaceTransforming],
    path: str,
) -> list[SolveStep]:
    code = _pull_first_md_block_from_completion(completion)
    if code is None:
        code = completion.completion_text
        if code.endswith("\n```"):
            code = code[: -len("\n```")]
    # Remove docstrings. This is because chat models might choose to add them or not
    code = _strip_prefix_comments(code, lang_spec)
    code_prefix = _strip_prefix_comments(code_prefix, lang_spec)
    code_prefix = code_prefix.rstrip()
    # Try to strip of repeated code from the prompt
    code = code[_get_new_content_start(code_prefix, code) :]
    if code.startswith("\n"):
        code = code[1:]
    code = format_return_val_for_node(code, insert_node, lang_spec)
    if insert_node.verb.stop_at_block_end:
        code = lang_spec.truncate_function_completion(code)
    if lang_spec.get_lang_md_name() == "python" and code.startswith('    """'):
        # Strip docstring
        code = code[code.find('"""') + 3 :]
        code = code[code.find('"""') + 3 :]
        if len(code) > 0 and code[0] == "\n":
            code = code[1:]
    return [SolveStep(path, insert_node.mark_id, code)]


def _strip_prefix_comments(
    code: str,
    lang_spec: LangSpec,
):
    if lang_spec.get_lang_md_name() == "python":
        comments_in_code = list(lang_spec.get_prefix_comments(code))
        for comment in comments_in_code:
            text = "    " + comment.doc_string_node.text.decode() + "\n"
            if text in code:
                code = code.replace(text, "")
    return code


def _pull_first_md_block_from_completion(
    completion: LmPrediction,
) -> str | None:
    """
    Pulls the first markdown block from a completion.
    Markdown blocks are ```...```.
    """
    text = completion.completion_text

    # Find the start of the code block
    start = text.find("```")
    if start == -1:
        return None

    # Find the end of the line after the starting ```
    # This is where the markdown language name is usually located
    end_of_language_name = text.find("\n", start + 3)
    if end_of_language_name == -1:
        # if there's no newline after the starting ```, then it's not a valid block
        return None

    # Now find the end of the code block
    end = text.find("```", end_of_language_name)
    if end == -1:
        # if there's no closing ```, then it's not a valid block
        return None

    # Return the text inside the block, excluding the markdown language name
    return text[end_of_language_name + 1 : end]


def _get_new_content_start(first_str: str, new_content: str) -> int:
    """
    Find the point in the new_content string where to slice so everything
    after is unique from the suffix of the first string.
    This is used to find the point where the new generation starts if
    there is possible repeating of the prompt.

    GPT-4 gen:

    Approach:
        We use the rolling hash technique to compute the hash values of the
        suffix of first_str and the prefix of new_content in linear time.
        While computing these hashes, we compare them at each step.
        If they are equal, it means that the current suffix of first_str
        is also a prefix of new_content. We remember this position.
        The last position at which the hashes were equal will give us the
        longest suffix of first_str that is also a prefix of new_content.

    Note:
    ----
        This function may return false positives (collisions) because two different strings
        might have the same hash, but it is very unlikely when we use a large prime modulus.

    Returns:
    -------
        An integer position value representing the starting point of 'new' content in new_content.

    """
    # TODO: some way of skipping comments

    # Initializing hashes and multipliers
    # m is a large prime number used as modulus to prevent overflow and hash collision
    # p is another prime number used as base of our polynomial hash function
    m, p = 1000000007, 31
    power, hash_first, hash_new = 1, 0, 0

    result = 0  # Initialize result position

    for i in range(len(first_str)):
        # Calculate ASCII value of char
        symbol_first = ord(first_str[-i - 1])
        symbol_new = ord(new_content[i]) if i < len(new_content) else 0

        # Compute the hash of the last (i+1) chars of first_str, and
        # first (i+1) chars of new_content using rolling hash technique.
        hash_first = (hash_first + symbol_first * power) % m
        hash_new = (hash_new * p + symbol_new) % m
        power = (power * p) % m

        # If hashes are equal, update result position
        if hash_first == hash_new:
            result = i + 1

    return result



import errno
import os
import select
import signal
import socket
import struct
import time

import dateutil.parser
import docker
import structlog
from docker import constants as docker_consts
from docker.errors import DockerException
from docker.types import Ulimit
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException
from urllib3.util.retry import Retry

from . import config, exceptions

logger = structlog.get_logger()

_DOCKER_CLIENTS = {}

#: Recoverable IO/OS Errors.
ERRNO_RECOVERABLE = (errno.EINTR, errno.EDEADLK, errno.EWOULDBLOCK)


def get_docker_client(
    base_url=None,
    retry_read=config.DOCKER_MAX_READ_RETRIES,
    retry_status_forcelist=(500,),
):
    client_key = (retry_read, retry_status_forcelist)
    if client_key not in _DOCKER_CLIENTS:
        client = docker.DockerClient(
            base_url=base_url or config.DOCKER_URL, timeout=config.DOCKER_TIMEOUT
        )
        retries = Retry(
            total=config.DOCKER_MAX_TOTAL_RETRIES,
            connect=config.DOCKER_MAX_CONNECT_RETRIES,
            read=retry_read,
            allowed_methods=None,
            status_forcelist=retry_status_forcelist,
            backoff_factor=config.DOCKER_BACKOFF_FACTOR,
            raise_on_status=False,
        )
        http_adapter = HTTPAdapter(max_retries=retries)
        client.api.mount("http://", http_adapter)
        _DOCKER_CLIENTS[client_key] = client
    return _DOCKER_CLIENTS[client_key]


def inspect_container_node(container):
    # 404 No such container may be returned when TimeoutError occurs
    # on container creation.
    docker_client = get_docker_client(retry_status_forcelist=(404, 500))
    try:
        container = docker_client.containers.get(container.id)
    except (RequestException, DockerException) as e:
        logger.exception("Failed to get the container", container=container)
        raise exceptions.DockerError(str(e))
    if "Node" not in container.attrs:
        # Remote Docker side is not a Docker Swarm cluster
        return None
    return container.attrs["Node"]["Name"]


def inspect_exited_container_state(container):
    try:
        container.reload()
    except (RequestException, DockerException) as e:
        logger.exception(
            "Failed to load the container from the Docker engine", container=container
        )
        raise exceptions.DockerError(str(e))
    started_at = dateutil.parser.parse(container.attrs["State"]["StartedAt"])
    finished_at = dateutil.parser.parse(container.attrs["State"]["FinishedAt"])
    duration = finished_at - started_at
    duration_seconds = duration.total_seconds()
    if duration_seconds < 0:
        duration_seconds = -1
    return {
        "exit_code": container.attrs["State"]["ExitCode"],
        "duration": duration_seconds,
        "oom_killed": container.attrs["State"].get("OOMKilled", False),
    }


def demultiplex_docker_stream(data):
    """
    Demultiplex the raw docker stream into separate stdout and stderr streams.

    Docker multiplexes streams together when there is no PTY attached, by
    sending an 8-byte header, followed by a chunk of data.

    The first 4 bytes of the header denote the stream from which the data came
    (i.e. 0x01 = stdout, 0x02 = stderr). Only the first byte of these initial 4
    bytes is used.

    The next 4 bytes indicate the length of the following chunk of data as an
    integer in big endian format. This much data must be consumed before the
    next 8-byte header is read.

    Docs: https://docs.docker.com/engine/api/v1.24/#attach-to-a-container

    :param bytes data: A raw stream data.
    :return: A tuple `(stdout, stderr)` of bytes objects.
    """
    data_length = len(data)
    stdout_chunks = []
    stderr_chunks = []
    walker = 0
    while data_length - walker >= 8:
        header = data[walker : walker + docker_consts.STREAM_HEADER_SIZE_BYTES]
        stream_type, length = struct.unpack_from(">BxxxL", header)
        start = walker + docker_consts.STREAM_HEADER_SIZE_BYTES
        end = start + length
        walker = end
        if stream_type == 1:
            stdout_chunks.append(data[start:end])
        elif stream_type == 2:
            stderr_chunks.append(data[start:end])
    return b"".join(stdout_chunks), b"".join(stderr_chunks)


def _socket_read(sock, n=4096):
    """
    Read at most `n` bytes of data from the `sock` socket.

    :return: A bytes object or `None` at end of stream.
    """
    try:
        data = os.read(sock.fileno(), n)
    except OSError as e:
        if e.errno in ERRNO_RECOVERABLE:
            return b""
        raise e
    if data:
        return data


def _socket_write(sock, data):
    """
    Write as much data from the `data` buffer to the `sock` socket as possible.

    :return: The number of bytes sent.
    """
    try:
        return os.write(sock.fileno(), data)
    except OSError as e:
        if e.errno in ERRNO_RECOVERABLE:
            return 0
        raise e


def docker_communicate(container, stdin=None, start_container=True, timeout=None):
    """
    Interact with the container: Start it if required. Send data to stdin.
    Read data from stdout and stderr, until end-of-file is reached.

    :param Container container: A container to interact with.
    :param bytes stdin: The data to be sent to the standard input of the
                        container, or `None`, if no data should be sent.
    :param bool start_container: Whether to start the container after
                                 attaching to it.
    :param int timeout: Time in seconds to wait for the container to terminate,
        or `None` to make it unlimited.

    :return: A tuple `(stdout, stderr)` of bytes objects.

    :raise TimeoutError: If the container does not terminate after `timeout`
                         seconds. The container is not killed automatically.
    :raise RequestException, DockerException, OSError: If an error occurred
        with the underlying docker system.
    """
    # Retry on 'No such container' since it may happen when the attach/start
    # is called immediately after the container is created.
    docker_client = get_docker_client(retry_status_forcelist=(404, 500))
    log = logger.bind(container=container)
    params = {
        # Attach to stdin even if there is nothing to send to it to be able
        # to properly close it (stdin of the container is always open).
        "stdin": 1,
        "stdout": 1,
        "stderr": 1,
        "stream": 1,
        "logs": 0,
    }
    sock = docker_client.api.attach_socket(container.id, params=params)
    sock._sock.setblocking(False)  # Make socket non-blocking
    log.info(
        "Attached to the container", params=params, fd=sock.fileno(), timeout=timeout
    )
    if not stdin:
        log.debug("There is no input data. Shut down the write half " "of the socket.")
        sock._sock.shutdown(socket.SHUT_WR)
    if start_container:
        container.start()
        log.info("Container started")

    stream_data = b""
    start_time = time.time()
    while timeout is None or time.time() - start_time < timeout:
        read_ready, write_ready, _ = select.select([sock], [sock], [], 1)
        is_io_active = False
        if read_ready:
            is_io_active = True
            try:
                data = _socket_read(sock)
            except ConnectionResetError:
                log.warning(
                    "Connection reset caught on reading the container "
                    "output stream. Break communication"
                )
                break
            if data is None:
                log.debug("Container output reached EOF. Closing the socket")
                break
            stream_data += data

        if write_ready and stdin:
            is_io_active = True
            try:
                written = _socket_write(sock, stdin)
            except BrokenPipeError:
                # Broken pipe may happen when a container terminates quickly
                # (e.g. OOM Killer) and docker manages to close the socket
                # almost immediately before we're trying to write to stdin.
                log.warning(
                    "Broken pipe caught on writing to stdin. Break " "communication"
                )
                break
            stdin = stdin[written:]
            if not stdin:
                log.debug(
                    "All input data has been sent. Shut down the write "
                    "half of the socket."
                )
                sock._sock.shutdown(socket.SHUT_WR)

        if not is_io_active:
            # Save CPU time
            time.sleep(0.05)
    else:
        sock.close()
        raise TimeoutError("Container didn't terminate after timeout seconds")
    sock.close()
    return demultiplex_docker_stream(stream_data)


def filter_filenames(files):
    return [file["name"] for file in files if "name" in file]


def merge_limits_defaults(limits):
    if not limits:
        return config.DEFAULT_LIMITS
    is_realtime_specified = "realtime" in limits
    for limit_name, default_value in config.DEFAULT_LIMITS.items():
        if limit_name not in limits:
            limits[limit_name] = default_value
    if not is_realtime_specified:
        limits["realtime"] = limits["cputime"] * config.CPU_TO_REAL_TIME_FACTOR
    return limits


def create_ulimits(limits):
    ulimits = []
    if limits["cputime"]:
        cpu = limits["cputime"]
        ulimits.append(Ulimit(name="cpu", soft=cpu, hard=cpu))
    if "file_size" in limits:
        fsize = limits["file_size"]
        ulimits.append(Ulimit(name="fsize", soft=fsize, hard=fsize))
    return ulimits or None


def truncate_result(result):
    MAX_OUTPUT_LENGTH = 100
    truncated = {}
    for k, v in result.items():
        if k in ["stdout", "stderr"]:
            if len(v) > MAX_OUTPUT_LENGTH:
                v = v[:MAX_OUTPUT_LENGTH] + b" *** truncated ***"
        truncated[k] = v
    return truncated


def is_killed_by_sigkill_or_sigxcpu(status):
    return status - 128 in [signal.SIGKILL, signal.SIGXCPU]



def main():
    from pprint import pprint

    from datasets import load_dataset

    dataset = load_dataset("THUDM/humaneval-x", "java")
    for ex in dataset["test"]:
        pprint(ex)
        print("Prompt ---")
        print(ex["prompt"])
        print("Canonical_solution ---")
        print(ex["canonical_solution"])
        print("Test cases ---")
        print(ex["test"])
        break


if __name__ == "__main__":
    main()



import contextlib
import faulthandler
import io
import logging
import multiprocessing
import os
import platform
import signal
import tempfile
from pathlib import Path


def check_correctness(
    problem: dict,
    completion: str,
    timeout: float,
    completion_id: int | None = None,
) -> dict:
    """
    Evaluates the functional correctness of a completion by running the test
    suite provided in the problem.

    :param completion_id: an optional completion ID so we can match
        the results later even if execution finishes asynchronously.
    """

    def unsafe_execute():
        with create_tempdir():
            # These system calls are needed when cleaning up tempdir.
            import os
            import shutil

            rmtree = shutil.rmtree
            rmdir = os.rmdir
            chdir = os.chdir

            # Disable functionalities that can make destructive changes to the test.
            reliability_guard()

            # Construct the check program and run it.
            check_program = (
                problem["prompt"]
                + completion
                + "\n"
                + problem["test"]
                + "\n"
                + f"check({problem['entry_point']})"
            )

            try:
                exec_globals = {}
                with swallow_io(), time_limit(timeout):
                    # WARNING
                    # This program exists to execute untrusted model-generated code. Although
                    # it is highly unlikely that model-generated code will do something overtly
                    # malicious in response to this test suite, model-generated code may act
                    # destructively due to a lack of model capability or alignment.
                    # Users are strongly encouraged to sandbox this evaluation suite so that it
                    # does not perform destructive actions on their host or network. For more
                    # information on how OpenAI sandboxes its code, see the accompanying paper.
                    # Once you have read this disclaimer and taken appropriate precautions,
                    # uncomment the following line and proceed at your own risk
                    logging.info("RUN %s", check_program)
                    exec(check_program, exec_globals)
                result.append("passed")
            except TimeoutError:
                result.append("timed out")
            except BaseException as e:
                result.append(f"failed: {e}")

            # Needed for cleaning up.
            shutil.rmtree = rmtree
            os.rmdir = rmdir
            os.chdir = chdir

    manager = multiprocessing.Manager()
    result = manager.list()

    p = multiprocessing.Process(target=unsafe_execute)
    p.start()
    p.join(timeout=timeout + 1)
    if p.is_alive():
        p.kill()

    if not result:
        result.append("timed out")

    return {
        "task_id": problem["task_id"],
        "passed": result[0] == "passed",
        "result": result[0],
        "completion_id": completion_id,
    }


@contextlib.contextmanager
def time_limit(seconds: float):
    def signal_handler(signum, frame):
        msg = "Timed out!"
        raise TimeoutError(msg)

    signal.setitimer(signal.ITIMER_REAL, seconds)
    signal.signal(signal.SIGALRM, signal_handler)
    try:
        yield
    finally:
        signal.setitimer(signal.ITIMER_REAL, 0)


@contextlib.contextmanager
def swallow_io():
    stream = WriteOnlyStringIO()
    with (
        contextlib.redirect_stdout(stream),
        contextlib.redirect_stderr(
            stream,
        ),
        RedirectStdin(stream),
    ):
        yield


@contextlib.contextmanager
def create_tempdir():
    with tempfile.TemporaryDirectory() as dirname, chdir(dirname):
        yield dirname


class WriteOnlyStringIO(io.StringIO):
    """StringIO that throws an exception when it's read from."""

    def read(self, *args, **kwargs):
        raise OSError

    def readline(self, *args, **kwargs):
        raise OSError

    def readlines(self, *args, **kwargs):
        raise OSError

    def readable(self, *args, **kwargs):
        """Returns True if the IO object can be read."""
        return False


class RedirectStdin(contextlib._RedirectStream):  # type: ignore
    _stream = "stdin"


@contextlib.contextmanager
def chdir(root):
    if root == ".":
        yield
        return
    cwd = Path.cwd()
    os.chdir(root)
    try:
        yield
    finally:
        os.chdir(cwd)


def reliability_guard(maximum_memory_bytes: int | None = None):
    """
    Disables various destructive functions and prevents the generated code
    from interfering with the test (e.g. fork bomb, killing other processes,
    removing filesystem files, etc.).

    Warning:
    -------
    This function is NOT a security sandbox. Untrusted code, including, model-
    generated code, should not be blindly executed outside of one. See the
    Codex paper for more information about OpenAI's code sandbox, and proceed
    with caution.

    """
    if maximum_memory_bytes is not None:
        import resource

        resource.setrlimit(
            resource.RLIMIT_AS,
            (maximum_memory_bytes, maximum_memory_bytes),
        )
        resource.setrlimit(
            resource.RLIMIT_DATA,
            (maximum_memory_bytes, maximum_memory_bytes),
        )
        if platform.uname().system != "Darwin":
            resource.setrlimit(
                resource.RLIMIT_STACK,
                (maximum_memory_bytes, maximum_memory_bytes),
            )

    faulthandler.disable()

    import builtins

    builtins.exit = None
    builtins.quit = None
    builtins.open = None

    import os

    os.environ["OMP_NUM_THREADS"] = "1"

    os.kill = None
    os.system = None
    os.putenv = None
    os.remove = None
    os.removedirs = None
    os.rmdir = None
    os.fchdir = None
    os.setuid = None
    os.fork = None
    os.forkpty = None
    os.killpg = None
    os.rename = None
    os.renames = None
    os.truncate = None
    os.replace = None
    os.unlink = None
    os.fchmod = None
    os.fchown = None
    os.chmod = None
    os.chown = None
    os.chroot = None
    os.fchdir = None
    os.lchflags = None
    os.lchmod = None
    os.lchown = None
    os.getcwd = None
    os.chdir = None

    import shutil

    shutil.rmtree = None
    shutil.move = None
    shutil.chown = None

    import subprocess

    subprocess.Popen = None  # type: ignore

    __builtins__["help"] = None

    import sys

    sys.modules["ipdb"] = None
    sys.modules["joblib"] = None
    sys.modules["resource"] = None
    sys.modules["psutil"] = None
    sys.modules["tkinter"] = None



from synthegrator.code_problems import CodeProblem
from synthegrator.synthdatasets import yield_problems_from_name, DatasetName
from synthegrator.synthdatasets.human_eval import yield_human_eval
import pytest
import tempfile
from synthegrator.synthdatasets import DatasetName


def test_yield_from_dataset_names():
    data_by_name = list(yield_problems_from_name(
        DatasetName.humaneval,
        max_problems=5
    ))
    data_by_func = list(yield_human_eval(max_problems=5))
    assert data_by_name[0].working_directory == data_by_func[0].working_directory
    assert are_probs_eq(data_by_name[0], data_by_func[0])


@pytest.mark.parametrize(
    "dataset",
    set(DatasetName) - {DatasetName.dypy_line_completion, DatasetName.mbpp_sanatized},
)
def test_load_each_dataset(dataset):
    print(dataset)
    data = list(yield_problems_from_name(
        dataset,
        max_problems=1
    ))
    assert len(data) == 1
    assert data[0].dataset_name == dataset.get_base_collection_name()

try:
    import joblib
    JOBLIB_AVAILABLE = True
except ImportError:
    JOBLIB_AVAILABLE = False

# Create a decorator to skip tests that require joblib
joblib_required = pytest.mark.skipif(
    not JOBLIB_AVAILABLE,
    reason="Test requires joblib to be installed"
)

# Example usage
@joblib_required
def test_joblib_caching_with_dataset_names():
    """Test that joblib's Memory correctly caches with DatasetName attributes."""
    # Setup temporary cache directory
    cache_dir = tempfile.mkdtemp()
    memory = joblib.Memory(cache_dir, verbose=0)

    # Counter for actual function calls
    call_count = 0

    @memory.cache
    def cached_function(dataset, value):
        nonlocal call_count
        call_count += 1
        return (dataset.name if hasattr(dataset, 'name') else dataset, value)

    # Test 1: First call should increment counter, second should use cache
    cached_function(DatasetName.humaneval, 42)
    assert call_count == 1, "First call should execute function"

    cached_function(DatasetName.humaneval, 42)
    assert call_count == 1, "Second call should use cache (not increment counter)"

    # Test 2: Different parameter should increment counter
    cached_function(DatasetName.humaneval, 43)
    assert call_count == 2, "Call with different value should execute function"

    # Test 3: Different dataset should increment counter
    cached_function(DatasetName.mbpp, 42)
    assert call_count == 3, "Call with different dataset should execute function"

    # Test 4: Dataset with parent relationship
    cached_function(DatasetName.humaneval_plus, 42)
    assert call_count == 4, "Call with parent dataset should execute function"

    # Call again with same parent dataset
    cached_function(DatasetName.humaneval_plus, 42)
    assert call_count == 4, "Repeated call with parent dataset should use cache"

    # Test 5: String representation equivalence
    call_count_before = call_count
    cached_function("humaneval", 42)
    assert call_count > call_count_before, "String representation should be treated differently"

    # Cleanup
    memory.clear()


def are_probs_eq(a: CodeProblem, b: CodeProblem):
    return bool(a == b)



from collections.abc import Iterable
from dataclasses import dataclass

from tree_sitter import Language, Node, Tree
#from tree_sitter_languages import get_language
from tree_sitter_language_pack import get_language

from synthegrator.lang_specs.lang_spec import LangSpec, LsFunctionDef, PrefixComment
from synthegrator.sandboxing import JAVA_DEFAULT_DOCKER_ENV, DockerExecutionContext

JAVA_LANG_TS = get_language("java")


class JavaLangSpec(LangSpec):
    @staticmethod
    def get_tree_sitter_lang_name() -> str:
        return "java"

    def truncate_function_completion(self, completion: str) -> str:
        with_dummy = "public class Dummy {\n    public void dummy() {\n" + completion
        funcs = list(self.find_functions(with_dummy))
        if len(funcs) == 0:
            return completion
        func = funcs[0]
        body_src = func.get_body_src()
        if body_src.startswith("{") and not completion.startswith("{"):
            body_src = body_src[1:]
            if body_src.startswith("\n") and not completion.startswith("{\n"):
                body_src = body_src[1:]
        return body_src.rstrip()

    def get_method_name(self, method_src) -> str:
        raise NotImplementedError

    def get_lang_md_name(self) -> str:
        return "java"

    @staticmethod
    def split_before_last_method_name(code: str) -> tuple[str, str]:
        raise NotImplementedError

    @staticmethod
    def get_signature_text(method: str) -> str:
        raise NotImplementedError

    @staticmethod
    def get_default_docker_env() -> DockerExecutionContext:
        return JAVA_DEFAULT_DOCKER_ENV

    @classmethod
    def find_functions(
        cls,
        src: str | Tree,
        include_class_functions: bool = True,
        include_nested_functions: bool = True,
    ) -> Iterable["LsFunctionDefJava"]:
        """
        Finds all functions in the source code.
        :param include_class_functions: If true, will include functions defined inside classes.
        :param include_nested_functions: If true, will include functions defined inside other functions.
        """
        if not include_nested_functions:
            raise NotImplementedError
        if not include_class_functions:
            raise NotImplementedError
        tree = cls.get_tree_sitter_tree(src) if isinstance(src, str) else src
        captures = _java_func_def_query().captures(tree.root_node)
        for node in captures.get('function', []):
            yield LsFunctionDefJava(node=node, tree=tree)

    @classmethod
    def get_prefix_comments(cls, src: str | Tree) -> Iterable[PrefixComment]:
        raise NotImplementedError

    @staticmethod
    def get_tree_sitter_lang() -> Language:
        return JAVA_LANG_TS

    @classmethod
    def get_comment_line_start(cls):
        return "//"


@dataclass(frozen=True)
class LsFunctionDefJava(LsFunctionDef):
    """
    A wrapper around a function definition with useful functions common
    between languages.
    """

    def get_body_node(self) -> Node:
        """Returns the node that corresponds to the body of the function"""
        return self.node.children[-1]

    # @frozen_method_cache
    def get_body_src(
        self,
        include_body_indent: bool = True,
        include_trailing_new_line: bool = True,
    ) -> str:
        """
        Source of the method.

        These parameters don't really make sense for Java. Need to sort through that.
        Note that the starting { is included in the body.
        """
        self.get_body_node().text.decode()
        if not include_body_indent:
            raise NotImplementedError
        if not include_trailing_new_line:
            raise NotImplementedError
        return self.get_body_node().text.decode()

    def get_preceding_src_code(self):
        """Returns the source code before the function declaration"""
        raise NotImplementedError

    def get_src_code_after(self):
        """Returns the source code after the function"""
        raise NotImplementedError

    # @frozen_method_cache
    def get_function_name(self) -> str:
        """Returns the name identifier of the function"""
        raise NotImplementedError

    # @frozen_method_cache
    def get_declaration_str(self, include_body_indent: bool = False) -> str:
        r"""
        Returns the full declaration of the function (so the `def`, signature, args, and `:`

        :param include_body_indent: If false the return might have leading ident.
            For example "def foo():\n    pass" would return "def foo():\n    "
            with the indent. If false would return "def foo():\n"
        """
        raise NotImplementedError

    # @frozen_method_cache
    def get_full_function_src(self, zero_base_indent: bool = False) -> str:
        """Returns the entire source code of the function, optionally without leading indentation"""
        raise NotImplementedError

    # @frozen_method_cache
    def get_prefix_comment_str(self, as_plain_text: bool = False) -> str | None:
        """
        Gets the prefix comment (a docstring in Python)

        :param as_plain_text: Normally will include the language specific annotations
            (such as the triple quotes in Python). If true, will return the dedented plain text.
        """
        raise NotImplementedError


def _java_func_def_query():
    return JAVA_LANG_TS.query(
        """
        (method_declaration) @function
        """
    )



from unittest.mock import MagicMock

from synthegrator.problem_rendering_insertion_tags import (
    TaggedEditRenderer,
    TaggedEditResponseParser,
    _find_ids_and_values_in_tagged_answer,
    _get_mark_default_id,
    _marked_up_text_to_tagged_prompt_text,
)
from synthegrator.tests.non_dataset_tests.test_problem_rendering import (
    factor_problems,
)
from synthegrator.transformation_spec import (
    MarkElement,
    MarkText,
    StsInsert,
    StsInvisible,
    StsNormalText,
)

assert factor_problems, "Unused import detectors struggle with fixtures"


def test_tagged_prompt():
    markup = MarkElement(
        verb=StsNormalText(),
        children=[
            MarkText("foo", 0, 3),
            MarkElement(StsInsert(), [], 3, 3),
            MarkText("bar", 3, 6),
        ],
        start_char=0,
        end_char=6,
    ).freeze()
    out, tag_ids = _marked_up_text_to_tagged_prompt_text(
        markup,
        path_id="0",
        # path="solution.py",
    )
    edit_node: MarkElement = markup.children[1]
    assert out == f'foo<edit id="0_{edit_node.mark_id}"></edit>bar'
    assert tag_ids == [f"0_{edit_node.mark_id}"]


def test_tagged_prompt2():
    markup = MarkElement(
        verb=StsNormalText(),
        children=[
            MarkText("foo", 0, 3),
            MarkElement(StsInvisible(), [MarkText("hide", 3, 7)], 3, 7),
            MarkText("bar", 7, 10),
        ],
        start_char=0,
        end_char=6,
    ).freeze()
    out, _tag_ids = _marked_up_text_to_tagged_prompt_text(
        markup,
        path_id="0",
    )
    markup.children[1]
    assert out == "foobar"


def test_insert_prompt(factor_problems):
    problem = factor_problems[0]
    renderer = TaggedEditRenderer()
    render = renderer(problem, MagicMock())
    content = (
        "Complete the following method please\n"
        "@@ foo.py @@\n"
        "def double(x) -> int:\n"
        '    """\n'
        "    Write a function that doubles its input.\n"
        '    """\n'
        '<edit id="0_qkIO0"></edit>\n'
        "\n"
        "---\n"
        "Implement the correct changes in the above code in the <edit> tags. Please "
        'place each fix in a separate <edit_solve id="id"> tags.\n'
        "For example:\n"
        '<edit_solve id="id_example">new replacement for the text...</edit_solve>\n'
        "Answer:\n"
    )
    assert render.prompt.text == content


def test_insert_prompt_default_id(factor_problems):
    problem = factor_problems[0]
    default = _get_mark_default_id(problem)
    assert default == "0_qkIO0"


def test_insert_prompt_custom(factor_problems):
    problem = factor_problems[0]
    renderer = TaggedEditRenderer(
        tag_name_edit="buggy",
        tag_name_solve="fix",
        custom_closing_lines="Please fix the code.",
    )
    render = renderer(problem, MagicMock())
    content = (
        "Complete the following method please\n"
        "@@ foo.py @@\n"
        "def double(x) -> int:\n"
        '    """\n'
        "    Write a function that doubles its input.\n"
        '    """\n'
        '<buggy id="0_qkIO0"></buggy>\n'
        "\n"
        "---\n"
        "Please fix the code."
    )
    assert render.prompt.text == content
    response_parser = TaggedEditResponseParser()
    parse = response_parser.parse(
        render,
        MagicMock(
            completion_text='<fix id="0_qkIO0">    return x * 2</fix>',
        ),
        problem,
    )
    assert len(parse) == 1
    assert parse[0].path == "foo.py"
    assert parse[0].mark_id == "qkIO0"
    assert parse[0].value == "    return x * 2"


def test_find_ids_in_answer(factor_problems):
    val = _find_ids_and_values_in_tagged_answer(
        'foo<edit_solve id="0_qkIO0">bar</edit_solve>baz',
        default_id="default",
    )
    assert val == {"0_qkIO0": "bar"}


def test_find_ids_in_answer_dup(factor_problems):
    val = _find_ids_and_values_in_tagged_answer(
        '<edit id="0_qkIO0">foo</edit><edit_solve id="0_qkIO0">bar</edit>',
        default_id="default",
    )
    assert val == {"0_qkIO0": "bar"}


def test_find_ids_in_answer_many(factor_problems):
    val = _find_ids_and_values_in_tagged_answer(
        '<edit_solve id="a">foo</edit_solve><edit_solve id="0_qkIO0">bar</edit>',
        default_id="default",
    )
    assert val == {"a": "foo", "0_qkIO0": "bar"}


def test_find_ids_in_answer_stop(factor_problems):
    val = _find_ids_and_values_in_tagged_answer(
        'foo\n<edit_solve id="0_qkIO0">bar',
        default_id="default",
    )
    assert val == {"0_qkIO0": "bar"}


def test_find_ids_in_answer_default(factor_problems):
    val = _find_ids_and_values_in_tagged_answer("foo", default_id="default")
    assert val == {"default": "foo"}


def test_find_ids_in_answer_default(factor_problems):
    val = _find_ids_and_values_in_tagged_answer(
        "prestuff<edit_solve>foo</edit_solve>", default_id="defaultid"
    )
    assert val == {"defaultid": "foo"}


def test_find_ids_in_answer_default2(factor_problems):
    val = _find_ids_and_values_in_tagged_answer(
        "foo</edit_solve>bar", default_id="default"
    )
    assert val == {"default": "foo"}


def test_full_parser(factor_problems):
    parser = TaggedEditResponseParser()
    problem = factor_problems[0]
    renderer = TaggedEditRenderer()
    render = renderer(problem, MagicMock())
    solve_steps = parser.parse(
        render,
        MagicMock(
            completion_text='<edit_solve id="0_qkIO0">    return x * 2</edit_solve>',
        ),
        problem,
    )
    assert len(solve_steps) == 1
    assert solve_steps[0].path == "foo.py"
    assert solve_steps[0].mark_id == "qkIO0"
    assert solve_steps[0].value == "    return x * 2"



"""
Some functions that can help make common types of problems.
TODO: a lot of these are pretty unweildy. There should
    probably be something with a builder pattern to
    gradually add different parts of the problem.
"""

import dataclasses

from synthegrator.code_problems import (
    CodeProblem,
    CodeSolution,
    LmCodeSolution,
    TestCase,
)
from synthegrator.environments import CodeExecutionEnvironment, ProjectWorkingDirectory
from synthegrator.lang_specs.lang_spec import LangSpec
from synthegrator.lang_specs.lang_spec_python import (
    PythonLangSpec,
    assert_spec_is_python,
)
from synthegrator.memory_fs import ProjectDir
from synthegrator.sandboxing import DockerExecutionContext
from synthegrator.transformation_spec import (
    MarkElement,
    SolveStep,
    StsEditable,
    StsInsert,
    StsInvisible,
    StsPath,
    StsSelectorCharSpan,
    StsSelectorLineSpan,
    StsSpecStatement,
    TransformationSpec,
    markup_path,
)


def make_simple_method_completion_problem(
    prompt: str,
    signature: str,
    test_cases: list[TestCase],
    model_context: str = "",
    execution_context_hidden_from_model: str = "",
    target_path: str = "solution.py",
    known_solutions: list[str] | None = None,
    lang_spec: LangSpec | None = None,
    prompt_already_docstring: bool = False,
    dataset_name: str | None = None,
    problem_id: str | None = None,
    docker_env: DockerExecutionContext = None,
    override_instructions: str | None = None,
    base_indent: int = 4,
    after_method_src: str | None = None,
) -> CodeProblem:
    if lang_spec is None:
        lang_spec = PythonLangSpec()
    assert_spec_is_python(lang_spec)
    if "\r" in signature:
        msg = "Signature not expected to have carriage returns"
        raise ValueError(msg)
    if "\r" in prompt:
        msg = "Prompt not expected to have carriage returns"
        raise ValueError(msg)
    signature = signature.rstrip("\n")
    signature_last_line = signature.split("\n")[-1]
    line_delta = len(signature_last_line) - len(signature_last_line.lstrip())
    indentation = " " * (line_delta + 4)
    if not prompt_already_docstring:
        prompt = (
            indentation
            + '"""\n'
            + "\n".join(f"{indentation}{line}" for line in prompt.split("\n"))
            + f"\n{indentation}"
            + '"""'
        )
    body = (
        execution_context_hidden_from_model
        + model_context
        + signature
        + ("\n" if not signature.endswith("\n") else "")
        + prompt
        + ("\n" if not prompt.endswith("\n") else "")
    )
    post_context = None if (not after_method_src) else "\n" + after_method_src
    return _auto_regressive_insert_problem(
        body,
        override_instructions
        or (
            "Complete the following method. "
            "Give only the solution body, and do not repeat the signature or docstring."
        ),
        target_path,
        test_cases,
        (
            [(0, len(execution_context_hidden_from_model))]
            if execution_context_hidden_from_model
            else None
        ),
        dataset_name,
        problem_id,
        known_solutions,
        lang_spec,
        docker_env,
        base_indent,
        stop_at_block_end=True,
        post_context=post_context,
    )


def _solution_strings_to_solves(
    markup: MarkElement,
    solution_strings: list[str],
    target_path: str,
    problem: CodeProblem,
    lang_spec: LangSpec,
) -> list[CodeSolution]:
    solution_solves = []
    for sol in solution_strings:
        insert_nodes = [
            node
            for node in markup.depth_first_iterate()
            if isinstance(node, MarkElement) and isinstance(node.verb, StsInsert)
        ]
        if len(insert_nodes) != 1:
            msg = "Insert node length must be one"
            raise ValueError(msg)
        insert_node = insert_nodes[0]
        solution_solves.append(
            CodeSolution(
                solve_steps=[
                    SolveStep(
                        path=target_path,
                        mark_id=insert_node.mark_id,
                        value=sol,
                    ),
                ],
                problem=problem,
                lang_spec=lang_spec,
            ),
        )
    return solution_solves


def _auto_regressive_insert_problem(
    body: str,
    instructions: str,
    target_path: str,
    test_cases: list[TestCase],
    invisible_spans: list[tuple[int, int]] | None = None,
    dataset_name: str | None = None,
    problem_id: str | None = None,
    known_solutions: list[str] | None = None,
    lang_spec: LangSpec | None = None,
    docker_env: DockerExecutionContext = None,
    base_indent: int = 4,
    stop_at_block_end: bool = True,
    post_context: str | None = None,
) -> CodeProblem:
    """Makes a problem with a single insert at the end"""
    if lang_spec is None:
        lang_spec = PythonLangSpec()
    if known_solutions is not None and type(known_solutions) != list:
        msg = "known_solutions must be a list"
        raise TypeError(msg)

    if post_context is None:
        post_context = ""

    wd = ProjectWorkingDirectory(
        files=ProjectDir.construct_with_one_file(target_path, body + post_context),
    )
    steps = [
        StsSpecStatement(
            StsPath(target_path),
            StsSelectorCharSpan(len(body), len(body)),
            StsInsert(
                set_base_space_indent=base_indent,
                stop_at_block_end=stop_at_block_end,
            ),
        ),
    ]
    steps.extend(
        StsSpecStatement(None, StsSelectorCharSpan(start, end), StsInvisible())
        for start, end in (invisible_spans or [])
    )
    spec = TransformationSpec(tuple(steps))
    markup = markup_path(wd, target_path, spec)
    problem = CodeProblem(
        instructions=instructions,
        working_directory=wd,
        transformation_spec=spec,
        test_cases=test_cases,
        dataset_name=dataset_name,
        problem_id=problem_id,
        environment=CodeExecutionEnvironment(docker_env=docker_env),
        instructions_are_essential=False,
    )
    problem.known_solutions = _solution_strings_to_solves(
        markup,
        known_solutions or [],
        target_path,
        problem,
        lang_spec,
    )
    return problem


def make_simple_file_generation(
    prompt: str,
    test_cases: list[TestCase],
    model_context: str = "",
    execution_context_hidden_from_model: str = "",
    target_path: str = "solution.py",
    lang_spec: LangSpec | None = None,
    dataset_name: str | None = None,
    problem_id: str | None = None,
    known_solutions: list[str] | None = None,
    docker_env: DockerExecutionContext = None,
) -> CodeProblem:
    if lang_spec is None:
        lang_spec = PythonLangSpec()
    # TODO: We are currently just quickly porting from old method, but
    #    we should support specifying file generation with a spec statement
    assert_spec_is_python(lang_spec)
    prompt = '""""\n' + "    \n".join(line for line in prompt.split("\n")) + '\n"""'
    content = (
        execution_context_hidden_from_model
        + "\n"
        + model_context
        + "\n"
        + prompt
        + "\n"
    )
    return _auto_regressive_insert_problem(
        body=content,
        instructions=f"Write {lang_spec.get_lang_md_name()} code following the instructions below",
        target_path=target_path,  # Fixed: this was previously swapped with instructions
        test_cases=test_cases,
        invisible_spans=(
            [(0, len(execution_context_hidden_from_model))]
            if execution_context_hidden_from_model
            else None
        ),
        dataset_name=dataset_name,
        problem_id=problem_id,
        known_solutions=known_solutions,
        lang_spec=lang_spec,
        docker_env=docker_env,
        base_indent=0,
    )


def _estimate_base_indent(text: str) -> int:
    """Estimates the base indentation of a text"""
    left_ws_len = len(text) - len(text.lstrip())
    left_ws = text[:left_ws_len]
    left_ws.replace("\t", " " * 4)
    left_ws.replace("\n", "")
    left_ws.replace("\r", "")
    return len(left_ws)


def make_simple_line_edit_problem(
    code: str,
    edit_lines_span: tuple[int, int],
    test_cases: list[TestCase],
    invisible_spans: list[tuple[int, int]] | None = None,
    current_text_visible: bool = True,
    target_path: str = "solution.py",
    known_solutions: list[str] | None = None,
    lang_spec: LangSpec | None = None,
    dataset_name: str | None = None,
    problem_id: str | None = None,
    docker_env: DockerExecutionContext = None,
    override_instructions: str | None = None,
    cap_to_max_same_num_lines: bool = True,
) -> CodeProblem:
    if lang_spec is None:
        lang_spec = PythonLangSpec()
    # assert_spec_is_python(lang_spec)
    start_idx, end_idx = edit_lines_span
    base_indent = _estimate_base_indent(code.splitlines()[start_idx])
    edit_statement = StsSpecStatement(
        StsPath(target_path),
        StsSelectorLineSpan(
            start_idx,
            end_idx,
            include_trailing_newline=True,
        ),
        StsEditable(
            current_text_visible=current_text_visible,
            set_ending_text="\n",
            set_base_space_indent=base_indent,
            max_length_lines=(
                end_idx - start_idx if cap_to_max_same_num_lines else None
            ),
        ),
    )

    if invisible_spans is None:
        statements = (edit_statement,)
    else:
        invisible_statements = [
            StsSpecStatement(
                StsPath(target_path),
                StsSelectorLineSpan(invisible_span_start, invisible_span_end),
                StsInvisible(),
            )
            for invisible_span_start, invisible_span_end in invisible_spans
        ]
        statements = (*invisible_statements, edit_statement)

    problem = CodeProblem(
        instructions=override_instructions,  # or "Edit the marked lines",
        working_directory=ProjectWorkingDirectory(
            files=ProjectDir.construct_with_one_file(target_path, code),
        ),
        transformation_spec=TransformationSpec(statements),
        test_cases=test_cases,
        dataset_name=dataset_name,
        problem_id=problem_id,
        environment=CodeExecutionEnvironment(docker_env=docker_env),
        instructions_are_essential=False,
    )
    if (
        problem.transformation_spec.count_editing_statements() != 1
    ):  # TODO only one StsPlaceTransforming
        msg = "Only one editing statement allowed"
        raise ValueError(msg)

    problem.known_solutions = [
        make_solve_for_single_edit(problem, sol) for sol in (known_solutions or [])
    ]
    return problem


def make_solve_for_single_edit(problem, solution: str) -> CodeSolution:
    from synthegrator.problem_rendering import (
        render_single_edit_codeproblem_autoregressive,
    )

    render = render_single_edit_codeproblem_autoregressive(problem)
    return LmCodeSolution(
        solve_steps=(SolveStep(render.path, render.changed_element.mark_id, solution),),
        problem=problem,
        lang_spec=PythonLangSpec(),
        prompt=dataclasses.replace(render.prompt, max_tokens=500),
    )


def make_instruction_only_problem(
    instructions: str,
    insert_node: StsInsert,
    test_cases: list[TestCase],
    known_solutions: list[str],
    dataset_name: str | None = None,
    problem_id: str | None = None,
    target_file_name: str = "solution.py",
) -> CodeProblem:
    wd, transformation_spec, insert_node_element = _instruction_only_data(
        target_file_name,
        insert_node,
    )
    problem = CodeProblem(
        instructions=instructions,
        working_directory=wd,
        test_cases=test_cases,
        transformation_spec=transformation_spec,
        dataset_name=dataset_name,
        problem_id=problem_id,
    )
    problem.known_solutions = [
        CodeSolution(
            solve_steps=(
                SolveStep(
                    target_file_name,
                    insert_node_element.mark_id,
                    known_sol,
                ),
            ),
            problem=problem,
            lang_spec=PythonLangSpec(),
        )
        for known_sol in known_solutions
    ]
    return problem


_wd_cache = {}


def _instruction_only_data(
    target_file_name: str,
    insert_node: StsInsert,
) -> tuple[ProjectWorkingDirectory | None, TransformationSpec, MarkElement]:
    wd = _wd_cache.get(target_file_name)
    if wd is None:
        wd = ProjectWorkingDirectory(
            ProjectDir.construct_with_one_file(target_file_name, ""),
        )
        _wd_cache[target_file_name] = wd
    transformation_spec = TransformationSpec(
        (
            StsSpecStatement(
                path_selector=StsPath(target_file_name),
                text_selector=StsSelectorCharSpan(0, 0),
                verb=insert_node,
            ),
        ),
    )
    markup = markup_path(wd, target_file_name, spec=transformation_spec)
    insert_node_element = next(
        node
        for node in markup.depth_first_iterate()
        if isinstance(node, MarkElement) and isinstance(node.verb, StsInsert)
    )
    return wd, transformation_spec, insert_node_element


"""
TODO?
class ProblemBuilder:
    def __init__(self):
        self._files = {}

    def add_file(self, file_name: str, content: str):
        pass

    def build(self):
        return CodeProblem(

        )
"""



import pandas as pd
from lmwrapper.openai_wrapper import get_open_ai_lm

from synthegrator.code_solver import LmCodeSolverAutoRegressive
from synthegrator.df_converters import solution_evals_to_df
from synthegrator.execution_threading import solve_and_evaluate_problems
from synthegrator.synthdatasets.human_eval import yield_human_eval


def main():
    problems = list(yield_human_eval())[:10]
    lm = get_open_ai_lm()
    solver = LmCodeSolverAutoRegressive(lm, include_lm_response=True)
    evals = solve_and_evaluate_problems(
        solver,
        problems,
        max_threads_solve=8,
        max_threads_eval=8,
    )
    df = solution_evals_to_df(evals)
    # make pandas no max width
    with pd.option_context("display.max_colwidth", None):
        print(df.columns)
        print(df.solution__lm_prediction__completion_logprobs)
        print(
            list(
                zip(
                    df.solution__lm_prediction__completion_logprobs.iloc[0],
                    df.solution__lm_prediction__completion_tokens.iloc[0],
                    strict=False,
                ),
            ),
        )


if __name__ == "__main__":
    main()
